from lisatools.detector import EqualArmlengthOrbits
import numpy as np
from typing import Optional, Any
from copy import deepcopy

from few.waveform import GenerateEMRIWaveform

# imports
from fastlisaresponse import ResponseWrapper
from lisatools.detector import EqualArmlengthOrbits

default_response_kwargs = dict(
    t0=30000.0,
    order=25,
    tdi="1st generation",
    tdi_chan="AET",
    orbits=EqualArmlengthOrbits(),
)


class EMRITDIWaveform(ResponseWrapper):
    """Generate EMRI waveforms with the TDI LISA Response

    Args:
        T: Observation time in years.
        dt: Time cadence in seconds.
        emri_waveform_args: Arguments for :class:`GenerateEMRIWaveforms`.
        emri_waveform_kwargs: Keyword arguments for :class:`GenerateEMRIWaveforms`.
        response_kwargs: Keyword arguments for :class:`ResponseWrapper`.

    """

    def __init__(
        self,
        T: Optional[float] = 1.0,
        dt: Optional[float] = 10.0,
        emri_waveform_args: Optional[tuple] = ("FastSchwarzschildEccentricFlux",),
        emri_waveform_kwargs: Optional[dict] = {},
        response_kwargs: Optional[dict] = default_response_kwargs,
    ):
        # sky parameters in GenerateEMRIWaveform
        index_lambda = 8
        index_beta = 7

        for key in default_response_kwargs:
            response_kwargs[key] = response_kwargs.get(
                key, default_response_kwargs[key]
            )
        gen_wave = GenerateEMRIWaveform(
            *emri_waveform_args,
            sum_kwargs=dict(pad_output=True),
            **emri_waveform_kwargs
        )

        response_kwargs_in = deepcopy(response_kwargs)
        # parameters
        super(EMRITDIWaveform, self).__init__(
            gen_wave,
            T,
            dt,
            index_lambda,
            index_beta,
            flip_hx=True,  # set to True if waveform is h+ - ihx
            remove_sky_coords=False,
            is_ecliptic_latitude=False,
            remove_garbage=True,  # removes the beginning of the signal that has bad information
            **response_kwargs_in
        )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        __doc__ = super(EMRITDIWaveform, self).__call__.__doc__
        return super(EMRITDIWaveform, self).__call__(*args, **kwargs)



# -*- coding: utf-8 -*-

import numpy as np
from scipy import stats
import warnings

try:
    import cupy as xp

    gpu_available = True
except ModuleNotFoundError:
    import numpy as xp

    gpu_available = False

from eryn.moves import GroupStretchMove
from eryn.prior import ProbDistContainer
from eryn.utils.utility import groups_from_inds
from .gbmultipletryrj import GBMutlipleTryRJ

from ...diagnostic import inner_product


__all__ = ["GBGroupStretchMove"]


# MHMove needs to be to the left here to overwrite GBBruteRejectionRJ RJ proposal method
class GBGroupStretchMove(GroupStretchMove, GBMutlipleTryRJ):
    """Generate Revesible-Jump proposals for GBs with try-force rejection

    Will use gpu if template generator uses GPU.

    Args:
        priors (object): :class:`ProbDistContainer` object that has ``logpdf``
            and ``rvs`` methods.

    """

    def __init__(
        self,
        gb_args,
        gb_kwargs,
        start_ind_limit=10,
        *args,
        **kwargs
    ):

        self.name = "gbgroupstretch"
        self.start_ind_limit = start_ind_limit
        GBMutlipleTryRJ.__init__(self, *gb_args, **gb_kwargs)
        GroupStretchMove.__init__(self, *args, **kwargs)

    


import numpy as np
import cupy as xp
from copy import deepcopy

from eryn.moves import RedBlueMove, StretchMove
from eryn.state import State

from lisatools.sampling.moves.skymodehop import SkyMove
from bbhx.likelihood import NewHeterodynedLikelihood
from tqdm import tqdm


class MBHSpecialMove(RedBlueMove):
    def __init__(self, waveform_gen, fd, data_residuals, psd, num_repeats, transform_fn, mbh_priors, mbh_kwargs, moves, df, temperature_controls, **kwargs):

        RedBlueMove.__init__(self, **kwargs)
        self.fd = fd
        self.data_residuals = data_residuals
        self.psd = psd
        self.waveform_gen = waveform_gen
        self.num_repeats = num_repeats
        self.transform_fn = transform_fn
        self.mbh_priors = mbh_priors
        self.mbh_kwargs = mbh_kwargs
        moves_tmp = [move[0] for move in moves]
        move_weights = [move[1] for move in moves]
        self.moves = moves_tmp
        self.move_weights = move_weights
        self.df = df
        self.temperature_controls = temperature_controls

    def get_logl(self):
        return -1/2 * (4 * self.df * self.xp.sum((self.data_residuals[:2].conj() * self.data_residuals[:2]) / self.psd[:2], axis=-1)).get()

    def propose(self, model, state):

        new_state = deepcopy(state)

        # TODO: in TF, can we do multiple together?
        # TODO: switch to heterodyning

        ntemps, nwalkers, nleaves, ndim = new_state.branches["mbh"].shape

        assert len(self.temperature_controls) == nleaves

        temp_inds_base = np.repeat(np.arange(ntemps)[:, None], nwalkers, axis=-1)
        walker_inds_base = np.tile(np.arange(nwalkers), (ntemps, 1))

        start_leaf = np.random.randint(0, nleaves)
        for base_leaf in range(nleaves):
            leaf = (base_leaf + start_leaf) % nleaves

            temperature_control_here = self.temperature_controls[leaf]
            temperature_control_here.betas[:] = new_state.betas_all[leaf]

            # remove cold chain sources
            xp.get_default_memory_pool().free_all_blocks()
            removal_coords = new_state.branches["mbh"].coords[0, :, leaf]
            removal_coords_in = self.transform_fn.both_transforms(removal_coords)

            removal_waveforms = self.waveform_gen(*removal_coords_in.T, fill=True, freqs=self.fd, **self.mbh_kwargs).transpose(1, 0, 2)
            assert removal_waveforms.shape == self.data_residuals.shape

            # TODO: fix T channel 
            # d - h -> need to add removal waveforms
            # ll_tmp1 = (-1/2 * 4 * self.df * xp.sum(self.data_residuals[:2].conj() * self.data_residuals[:2] / self.psd[:2], axis=(0, 2)) - xp.sum(xp.log(xp.asarray(self.psd[:2])), axis=(0, 2))).get()

            ll_tmp2 = ((-1/2 * 4 * self.df * xp.sum(self.data_residuals[:2].conj() * self.data_residuals[:2] / self.psd[:2], axis=(0, 2))) - xp.sum(xp.log(xp.asarray(self.psd[:2])), axis=(0, 2))).get()

            self.data_residuals[:2] += removal_waveforms[:2]
            
            # ll_tmp3 = (-1/2 * 4 * self.df * xp.sum(self.data_residuals[:2].conj() * self.data_residuals[:2] / self.psd[:2], axis=(0, 2))) #  - xp.sum(xp.log(xp.asarray(self.psd[:2])), axis=(0, 2))).get()

            psd_term = -xp.sum(xp.log(xp.asarray(self.psd[:2])), axis=(0, 2)).get()

            keep_het = ll_tmp2.argmax().item()
            del removal_waveforms
            xp.get_default_memory_pool().free_all_blocks()
            data_index = xp.arange(removal_coords_in.shape[0], dtype=np.int32)
            noise_index = xp.arange(removal_coords_in.shape[0], dtype=np.int32)
            het_coords = np.tile(removal_coords_in[keep_het], (removal_coords_in.shape[0], 1))

            like_het = NewHeterodynedLikelihood(
                self.waveform_gen,
                self.fd,
                self.data_residuals.transpose(1, 0, 2).copy(),
                self.psd.transpose(1, 0, 2).copy(),
                het_coords,
                256,
                data_index=data_index,
                noise_index=noise_index,
                use_gpu=True,  # self.use_gpu,
            )
            
            ll_tmp = like_het.get_ll(
                removal_coords_in, 
                constants_index=data_index,
            ) + psd_term

            old_coords = new_state.branches["mbh"].coords[:, :, leaf].reshape(-1, ndim)
            old_coords_in = self.transform_fn.both_transforms(old_coords)
            data_index = walker_inds_base.astype(np.int32)
            noise_index = walker_inds_base.astype(np.int32)

            self.waveform_gen.d_d = ll_tmp  # xp.asarray(-2 * np.tile(ll_tmp2, (ntemps, 1)).flatten())

            del ll_tmp2
            xp.get_default_memory_pool().free_all_blocks()
            # d_d_store = self.waveform_gen.d_d.reshape(ntemps, nwalkers).get()

            data_index_in = xp.tile(xp.arange(nwalkers), (ntemps, 1)).flatten().astype(xp.int32)
            # TODO: fix this
            # prev_logl = self.waveform_gen.get_direct_ll(self.fd, self.data_residuals.flatten(), self.psd.flatten(), self.df, *old_coords_in.T, noise_index=noise_index, data_index=data_index, **self.mbh_kwargs).reshape((ntemps, nwalkers)).real.get()
            prev_logl = like_het.get_ll(
                old_coords_in, 
                constants_index=data_index_in,
            ).reshape((ntemps, nwalkers)).real + psd_term

            prev_logp = self.mbh_priors["mbh"].logpdf(old_coords).reshape((ntemps, nwalkers))

            prev_logP = temperature_control_here.compute_log_posterior_tempered(prev_logl, prev_logp)
            
            # fix this need to compute prev_logl for all walkers
            xp.get_default_memory_pool().free_all_blocks()
            for repeat in range(self.num_repeats):

                # pick move
                move_here = self.moves[model.random.choice(np.arange(len(self.moves)), p=self.move_weights)]

                # Split the ensemble in half and iterate over these two halves.
                accepted = np.zeros((ntemps, nwalkers), dtype=bool)
                all_inds = np.tile(np.arange(nwalkers), (ntemps, 1))
                inds = all_inds % self.nsplits
                if self.randomize_split:
                    [np.random.shuffle(x) for x in inds]

                # prepare accepted fraction
                accepted_here = np.zeros((ntemps, nwalkers), dtype=bool)
                for split in range(self.nsplits):
                    # get split information
                    S1 = inds == split
                    num_total_here = np.sum(inds == split)
                    nwalkers_here = np.sum(S1[0])

                    temp_inds_here = temp_inds_base[inds == split]
                    walker_inds_here = walker_inds_base[inds == split]

                    # prepare the sets for each model
                    # goes into the proposal as (ntemps * (nwalkers / subset size), nleaves_max, ndim)
                    sets = [
                        new_state.branches["mbh"].coords[inds == j][:, leaf].reshape(ntemps, -1, 1, ndim)
                        for j in range(self.nsplits)
                    ]

                    old_points = sets[split].reshape((ntemps, nwalkers_here, ndim))
                    
                    # setup s and c based on splits
                    s = {"mbh": sets[split]}
                    c = {"mbh": sets[:split] + sets[split + 1 :]}
                    
                    # Get the move-specific proposal.
                    if isinstance(move_here, StretchMove):
                        q, factors = move_here.get_proposal(
                            s, c, model.random
                        )

                    else:
                        q, factors = move_here.get_proposal(s, model.random)

                    new_points = q["mbh"].reshape((ntemps, nwalkers_here, ndim))

                    # Compute prior of the proposed position
                    # new_inds_prior is adjusted if product-space is used
                    logp = self.mbh_priors["mbh"].logpdf(new_points.reshape(-1, ndim))

                    new_points_in = self.transform_fn.both_transforms(new_points.reshape(-1, ndim)[~np.isinf(logp)])
                    
                    # Compute the lnprobs of the proposed position.
                    data_index = xp.asarray(walker_inds_here[~np.isinf(logp)].astype(np.int32))
                    # noise_index = walker_inds_here[~np.isinf(logp)].astype(np.int32)

                    # self.waveform_gen.d_d = xp.asarray(d_d_store[(temp_inds_here[~np.isinf(logp)], walker_inds_here[~np.isinf(logp)])])
                    
                    logl = np.full_like(logp, -1e300)

                    # logl[~np.isinf(logp)] = self.waveform_gen.get_direct_ll(self.fd, self.data_residuals.flatten(), self.psd.flatten(), self.df, *new_points_in.T, noise_index=noise_index, data_index=data_index, **self.mbh_kwargs).real.get()
                    logl[~np.isinf(logp)] = like_het.get_ll(
                        new_points_in, 
                        constants_index=data_index,
                    ) + psd_term[data_index.get()]
                    logl = logl.reshape(ntemps, nwalkers_here)
                    
                    logp = logp.reshape(ntemps, nwalkers_here)
                    prev_logp_here = prev_logp[inds == split].reshape(ntemps, nwalkers_here)

                    prev_logl_here = prev_logl[inds == split].reshape(ntemps, nwalkers_here)

                    prev_logP_here = temperature_control_here.compute_log_posterior_tempered(prev_logl_here, prev_logp_here)
                    logP = temperature_control_here.compute_log_posterior_tempered(logl, logp)

                    lnpdiff = factors + logP - prev_logP_here

                    keep = lnpdiff > np.log(model.random.rand(ntemps, nwalkers_here))

                    temp_inds_update = temp_inds_here[keep.flatten()]
                    walker_inds_update = walker_inds_here[keep.flatten()]

                    accepted[(temp_inds_update, walker_inds_update)] = True

                    # update state informatoin
                    new_state.branches["mbh"].coords[(temp_inds_update, walker_inds_update, np.full_like(walker_inds_update, leaf))] = new_points[keep].reshape(len(temp_inds_update), ndim)

                    prev_logl[(temp_inds_update, walker_inds_update)] = logl[keep].flatten()
                    prev_logp[(temp_inds_update, walker_inds_update)] = logp[keep].flatten()
                    prev_logP[(temp_inds_update, walker_inds_update)] = logP[keep].flatten()

                # acceptance tracking
                self.accepted += accepted
                # print(self.accepted[0])
                self.num_proposals += 1

                # TODO: include PSD likelihood in swaps?
                # temperature swaps
                # make swaps
                coords_for_swap = {"mbh": new_state.branches_coords["mbh"][:, :, leaf].copy()[:, :, None]}
                
                # TODO: check permute make sure it is okay
                coords_for_swap, prev_logP, prev_logl, prev_logp, inds, blobs, supps, branch_supps = temperature_control_here.temperature_swaps(
                    coords_for_swap,
                    prev_logP.copy(),
                    prev_logl.copy(),
                    prev_logp.copy(),
                    branch_supps={"mbh":None}
                )

                temperature_control_here.adapt_temps()

                new_state.branches_coords["mbh"][:, :, leaf] = coords_for_swap["mbh"][:, :, 0]

            # ll_tmp1 = -1/2 * 4 * self.df * xp.sum(self.data_residuals[:2].conj() * self.data_residuals[:2] / self.psd[:2], axis=(0, 2)).get()

            # add back cold chain sources
            xp.get_default_memory_pool().free_all_blocks()

            add_coords = new_state.branches["mbh"].coords[0, :, leaf]
            add_coords_in = self.transform_fn.both_transforms(add_coords)

            add_waveforms = self.waveform_gen(*add_coords_in.T, fill=True, freqs=self.fd, **self.mbh_kwargs).transpose(1, 0, 2)
            assert add_waveforms.shape == self.data_residuals.shape

            # d - h -> need to subtract added waveforms
            self.data_residuals[:2] -= add_waveforms[:2]

            del like_het
            del add_waveforms
            xp.get_default_memory_pool().free_all_blocks()
            # ll_tmp2 = -1/2 * 4 * self.df * xp.sum(self.data_residuals[:2].conj() * self.data_residuals[:2] / self.psd[:2], axis=(0, 2)).get()

            # read out all betas from temperature controls
            new_state.betas_all[leaf] = temperature_control_here.betas[:]
            # print(leaf)

            # ll_tmp2 = -1/2 * 4 * self.df * xp.sum(self.data_residuals[:2].conj() * self.data_residuals[:2] / self.psd[:2], axis=(0, 2)).get()

        # udpate at the end
        # new_state.log_like[(temp_inds_update, walker_inds_update)] = logl.flatten()
        # new_state.log_prior[(temp_inds_update, walker_inds_update)] = logp.flatten()

        current_ll = (-1/2 * 4 * self.df * xp.sum(self.data_residuals[:2].conj() * self.data_residuals[:2] / self.psd[:2], axis=(0, 2)) - xp.sum(xp.log(xp.asarray(self.psd[:2])), axis=(0, 2))).get()
        xp.get_default_memory_pool().free_all_blocks()
        # TODO: add check with last used logl

        current_lp = self.mbh_priors["mbh"].logpdf(new_state.branches["mbh"].coords[0, :, :].reshape(-1, ndim)).reshape(new_state.branches["mbh"].shape[1:-1]).sum(axis=-1)

        new_state.log_like[0] = current_ll
        new_state.log_prior[0] = current_lp
        xp.get_default_memory_pool().free_all_blocks()
        if not hasattr(self, "best_last_ll"):
            self.best_last_ll = current_ll.max()
            self.low_last_ll = current_ll.min()
        # print("mbh", self.best_last_ll, current_ll.max(), current_ll.max() - self.best_last_ll)
        # print(current_ll.max(), self.best_last_ll, current_ll.min(), self.low_last_ll)
        self.best_last_ll = current_ll.max()
        self.low_last_ll = current_ll.min()
        self.temperature_control.swaps_accepted = self.temperature_controls[0].swaps_accepted
        return new_state, accepted


from multiprocessing.sharedctypes import Value
from types import ModuleType, NoneType
from typing import Tuple
import numpy as np

# from ..sensitivity import get_sensitivity

try:
    import cupy as cp

except (ModuleNotFoundError, ImportError):
    import numpy as cp

    pass


def get_array_module(arr: np.ndarray | cp.ndarray) -> ModuleType:
    """Return array library of an array (np/cp).

    Args:
        arr: Numpy or Cupy array.

    """
    if isinstance(arr, np.ndarray):
        return np
    elif isinstance(arr, cp.ndarray):
        return cp
    else:
        raise ValueError("arr must be a numpy or cupy array.")


def generate_noise_fd(freqs, df, *sensitivity_args, func=None, **sensitivity_kwargs):
    if func is None:
        func = get_sensitivity

    norm = 0.5 * (1.0 / df) ** 0.5
    psd = func(freqs, *sensitivity_args, **sensitivity_kwargs)
    noise_to_add = psd ** (1 / 2) * (
        np.random.normal(0, norm, len(freqs))
        + 1j * np.random.normal(0, norm, len(freqs))
    )
    return noise_to_add


def AET(
    X: float | np.ndarray, Y: float | np.ndarray, Z: float | np.ndarray
) -> Tuple[float | np.ndarray, float | np.ndarray, float | np.ndarray]:
    """Transform to AET from XYZ

    .. math::

        A = (Z - X) / \\sqrt(2)

    .. math::

        E = (X - 2Y + Z) / \\sqrt(6)

    .. math::

        T = (X + Y + Z) / \\sqrt(3)

    Args:
        X: X-channel information.
        Y: Y-channel information.
        Z: Z-channel information.

    Returns:
        A, E, T Channels.

    """
    return (
        (Z - X) / np.sqrt(2.0),
        (X - 2.0 * Y + Z) / np.sqrt(6.0),
        (X + Y + Z) / np.sqrt(3.0),
    )


def searchsorted2d_vec(a, b, xp=None, gpu=None, **kwargs):
    if xp is None:
        xp = np
    else:
        try:
            xp.cuda.runtime.setDevice(gpu)
        except AttributeError:
            pass

    m, n = a.shape
    max_num = xp.maximum(a.max() - a.min(), b.max() - b.min()) + 1
    r = max_num * xp.arange(a.shape[0])[:, None]
    p = xp.searchsorted((a + r).ravel(), (b + r).ravel(), **kwargs).reshape(m, -1)

    out = p - n * (xp.arange(m)[:, None])
    try:
        xp.cuda.runtime.deviceSynchronize()
    except AttributeError:
        pass

    return out


def get_groups_from_band_structure(
    f0, band_edges, f0_2=None, xp=None, num_groups_base=3, fix_f_test=None, inds=None
):
    if num_groups_base not in [2, 3, 4]:
        raise ValueError("num_groups_base must be 2 or 3 or 4.")
    if xp is None:
        xp = np

    else:
        try:
            xp.cuda.runtime.setDevice(xp.cuda.runtime.getDevice())

        except AttributeError:
            # it is numpy
            pass

    if not isinstance(f0, xp.ndarray) or not isinstance(band_edges, xp.ndarray):
        raise TypeError(
            "f0 and band_edges must be xp.ndarray with xp as numpy or cupy as given by the xp kwarg."
        )

    shape = f0.shape

    # remove any above or below bands
    bad = (f0 < band_edges.min()) | (f0 > band_edges.max())

    band_indices = xp.searchsorted(band_edges, f0.flatten()).reshape(shape) - 1

    # sort the bands in, but keep places with inds_band_indices
    band_indices_sorted = xp.sort(band_indices, axis=-1)
    inds_band_indices = xp.argsort(band_indices, axis=-1)

    if f0_2 is not None:
        assert f0_2.shape == f0.shape
        band_indices_2 = xp.searchsorted(band_edges, f0_2.flatten()).reshape(shape) - 1
        band_indices_2_sorted = xp.take_along_axis(
            band_indices_2, inds_band_indices, axis=-1
        )

        # very important: ensures the proposed new point is not further than 1 band away.
        diff = 1 if num_groups_base > 2 else 0
        keep = (
            np.abs(band_indices_2_sorted.flatten() - band_indices_sorted.flatten())
            <= diff
        )
        if fix_f_test is not None:
            keep[fix_f_test.flatten()] = False
        remove = ~keep

    else:
        keep = np.ones(np.prod(band_indices_sorted.shape), dtype=bool)

    # temperature index associated with each band
    temp_inds = xp.repeat(
        xp.arange(band_indices_sorted.shape[0]), np.prod(band_indices_sorted.shape[1:])
    )[
        keep
    ]  # .reshape(shape)

    # walker index associated with each band
    walker_inds = (
        xp.tile(
            xp.arange(band_indices_sorted.shape[1]),
            (band_indices_sorted.shape[0], band_indices_sorted.shape[2], 1),
        )
        .transpose((0, 2, 1))
        .flatten()[keep]
    )

    if f0_2 is not None:
        temp_inds_remove = xp.repeat(
            xp.arange(band_indices_sorted.shape[0]),
            np.prod(band_indices_sorted.shape[1:]),
        )[
            remove
        ]  # .reshape(shape)

        # walker index associated with each band
        walker_inds_remove = (
            xp.tile(
                xp.arange(band_indices_sorted.shape[1]),
                (band_indices_sorted.shape[0], band_indices_sorted.shape[2], 1),
            )
            .transpose((0, 2, 1))
            .flatten()[remove]
        )
        inds_band_indices_remove = inds_band_indices.flatten()[remove]

    # special indexing method
    band_indices_sorted_special = (
        band_indices_sorted.flatten()[keep]
        + int(1e12) * temp_inds
        + int(1e6) * walker_inds
    )

    # get the unique special indicators
    (
        unique_special,
        unique_special_start_inds,
        unique_special_reverse,
        unique_special_counts,
    ) = np.unique(
        band_indices_sorted_special,
        return_index=True,
        return_inverse=True,
        return_counts=True,
    )

    # this basically makes mini arange setups for each band
    # the added_contribution for the first unique band index is removed
    added_contribution = xp.arange(band_indices_sorted_special.shape[0])

    # gets the groups
    combined = band_indices_sorted_special + added_contribution
    groups = (
        combined - (combined[unique_special_start_inds])[unique_special_reverse]
    )  # .reshape(shape)

    groups_even_odd_tmp = xp.asarray(
        [
            (num_groups_base * groups + i)
            * (band_indices_sorted.flatten()[keep] % num_groups_base == i)
            for i in range(num_groups_base)
        ]
    )
    groups_even_odd = xp.sum(groups_even_odd_tmp, axis=0)

    groups_out = -2 * xp.ones_like(f0, dtype=int)
    groups_out[
        (temp_inds, walker_inds, inds_band_indices.flatten()[keep])
    ] = groups_even_odd

    groups_out[bad] = -1

    """if f0_2 is not None and not np.all(keep):
        fix = (temp_inds_remove, walker_inds_remove, inds_band_indices_remove)
        fix_2 = band_indices_2[fix]
        fix_1 = band_indices[fix]"""

    return groups_out



import warnings
from types import ModuleType, NoneType
from typing import Optional, Any, Tuple, List

from numpy.typing import ArrayLike
import matplotlib.pyplot as plt

from eryn.utils import TransformContainer

import numpy as np

try:
    import cupy as cp

except (ModuleNotFoundError, ImportError):
    import numpy as cp

    pass

from .sensitivity import get_sensitivity, SensitivityMatrix
from .datacontainer import DataResidualArray
from .utils.utility import get_array_module


def inner_product(
    sig1: np.ndarray | list | DataResidualArray,
    sig2: np.ndarray | list | DataResidualArray,
    dt: Optional[float] = None,
    df: Optional[float] = None,
    f_arr: Optional[float] = None,
    psd: Optional[str | NoneType | np.ndarray | SensitivityMatrix] = "LISASens",
    psd_args: Optional[tuple] = (),
    psd_kwargs: Optional[dict] = {},
    normalize: Optional[bool | str] = False,
    complex: Optional[bool] = False,
) -> float | complex:
    """Compute the inner product between two signals weighted by a psd.

    The inner product between time series :math:`a(t)` and :math:`b(t)` is

    .. math::

        \langle a | b \\rangle = 2\int_{f_\\text{min}}^{f_\\text{max}} \\frac{\\tilde{a}(f)^*\\tilde{b}(f) + \\tilde{a}(f)\\tilde{b}(f)^*}{S_n(f)} df\ \ ,

    where :math:`\\tilde{a}(f)` is the Fourier transform of :math:`a(t)` and :math:`S_n(f)` is the one-sided Power Spectral Density of the noise.

    The inner product can be left complex using the ``complex`` kwarg.

    **GPU Capability**: Pass CuPy arrays rather than NumPy arrays.

    Args:
        sig1: First signal to use for the inner product.
            Can be time-domain or frequency-domain.
            Must be 1D ``np.ndarray``, list of 1D ``np.ndarray``s, or 2D ``np.ndarray``
            across channels with shape ``(nchannels, data length)``.
        sig2: Second signal to use for the inner product.
            Can be time-domain or frequency-domain.
            Must be 1D ``np.ndarray``, list of 1D ``np.ndarray``s, or 2D ``np.ndarray``
            across channels with shape ``(nchannels, data length)``.
        dt: Time step in seconds. If provided, assumes time-domain signals.
        df: Constant frequency spacing. This will assume a frequency domain signal with constant frequency spacing.
        f_arr: Array of specific frequencies at which the signal is given.
        psd: Indicator of what psd to use. If a ``str``, this will be passed as the ``sens_fn`` kwarg to :func:`get_sensitivity`.
            If ``None``, it will be an array of ones. Or, you can pass a 1D ``np.ndarray`` of psd values that must be the same length
            as the frequency domain signals.
        psd_args: Arguments to pass to the psd function if ``type(psd) == str``.
        psd_kwargs: Keyword arguments to pass to the psd function if ``type(psd) == str``.
        normalize: Normalize the inner product. If ``True``, it will normalize the square root of the product of individual signal inner products.
            You can also pass ``"sig1"`` or ``"sig2"`` to normalize with respect to one signal.
        complex: If ``True``, return the complex value of the inner product rather than just its real-valued part.

    Returns:
        Inner product value.

    """
    # initial input checks and setup
    sig1 = DataResidualArray(sig1, dt=dt, f_arr=f_arr, df=df)
    sig2 = DataResidualArray(sig2, dt=dt, f_arr=f_arr, df=df)

    if sig1.nchannels != sig2.nchannels:
        raise ValueError(
            f"Signal 1 has {sig1.nchannels} channels. Signal 2 has {sig2.nchannels} channels. Must be the same."
        )

    xp = get_array_module(sig1[0])

    # checks
    for i in range(sig1.nchannels):
        if not type(sig1[0]) == type(sig1[i]) and type(sig1[0]) == type(sig2[i]):
            raise ValueError(
                "Array in sig1, index 0 sets array module. Not all arrays match that module type (Numpy or Cupy)"
            )

    if sig1.data_length != sig2.data_length:
        raise ValueError(
            "The two signals are two different lengths. Must be the same length."
        )

    freqs = sig1.f_arr

    # get psd weighting
    if not isinstance(psd, SensitivityMatrix):
        psd = SensitivityMatrix(freqs, [psd], *psd_args, **psd_kwargs)

    operational_sets = []

    if psd.ndim == 3:
        assert psd.shape[0] == psd.shape[1] == sig1.shape[0] == sig2.shape[0]

        for i in range(psd.shape[0]):
            for j in range(i, psd.shape[1]):
                factor = 1.0 if i == j else 2.0
                operational_sets.append(
                    dict(factor=factor, sig1_ind=i, sig2_ind=j, psd_ind=(i, j))
                )

    elif psd.ndim == 2 and psd.shape[0] > 1:
        assert psd.shape[0] == sig1.shape[0] == sig2.shape[0]
        for i in range(psd.shape[0]):
            operational_sets.append(dict(factor=1.0, sig1_ind=i, sig2_ind=i, psd_ind=i))

    elif psd.ndim == 2 and psd.shape[0] == 1:
        for i in range(sig1.shape[0]):
            operational_sets.append(dict(factor=1.0, sig1_ind=i, sig2_ind=i, psd_ind=0))

    else:
        raise ValueError("# TODO")

    if complex:
        func = lambda x: x
    else:
        func = xp.real

    # initialize
    out = 0.0
    x = freqs

    # account for hp and hx if included in time domain signal
    for op_set in operational_sets:
        factor = op_set["factor"]

        temp1 = sig1[op_set["sig1_ind"]]
        temp2 = sig2[op_set["sig2_ind"]]
        psd_tmp = psd[op_set["psd_ind"]]

        ind_start = 1 if np.isnan(psd_tmp[0]) else 0

        y = (
            func(temp1[ind_start:].conj() * temp2[ind_start:]) / psd_tmp[ind_start:]
        )  # assumes right summation rule
        # df is sunk into trapz
        tmp_out = factor * 4 * xp.trapz(y, x=x[ind_start:])
        out += tmp_out

    # normalize the inner produce
    normalization_value = 1.0
    if normalize is True:
        norm1 = inner_product(
            sig1,
            sig1,
            psd=psd,
            normalize=False,
        )
        norm2 = inner_product(
            sig2,
            sig2,
            psd=psd,
            normalize=False,
        )

        normalization_value = np.sqrt(norm1 * norm2)

    elif isinstance(normalize, str):
        if normalize == "sig1":
            sig_to_normalize = sig1

        elif normalize == "sig2":
            sig_to_normalize = sig2

        else:
            raise ValueError(
                "If normalizing with respect to sig1 or sig2, normalize kwarg must either be 'sig1' or 'sig2'."
            )

        normalization_value = inner_product(
            sig_to_normalize,
            sig_to_normalize,
            psd=psd,
            normalize=False,
        )

    elif normalize is not False:
        raise ValueError("Normalize must be True, False, 'sig1', or 'sig2'.")

    out /= normalization_value
    return out


def residual_source_likelihood_term(
    data_res_arr: DataResidualArray, **kwargs: dict
) -> float | complex:
    """Calculate the source term in the Likelihood for a data residual (d - h).

    The source term in the likelihood is given by,

    .. math::

        \\log{\\mathcal{L}}_\\text{src} = -\\frac{1}{2}\\langle \\vec{d} - \\vec{h} | \\vec{d} - \\vec{h}\\rangle.

    Args:
        data_res_arr: Data residual.
        **kwargs: Keyword arguments to pass to :func:`inner_product`.

    Returns:
        Source term Likelihood value.

    """
    kwargs["normalize"] = False
    ip_val = inner_product(data_res_arr, data_res_arr, **kwargs)
    return -1 / 2.0 * ip_val


def noise_likelihood_term(psd: SensitivityMatrix) -> float:
    """Calculate the noise term in the Likelihood.

    The noise term in the likelihood is given by,

    .. math::

        \\log{\\mathcal{L}}_n = -\\sum \\log{\\vec{S}_n}.

    Args:
        psd: Sensitivity information.

    Returns:
        Noise term Likelihood value.

    """
    fix = np.isnan(psd[:]) | np.isinf(psd[:])
    assert np.sum(fix) == np.prod(psd.shape[:-1]) or np.sum(fix) == 0
    nl_val = -1.0 / 2.0 * np.sum(np.log(psd[~fix]))
    return nl_val


def residual_full_source_and_noise_likelihood(
    data_res_arr: DataResidualArray,
    psd: str | NoneType | np.ndarray | SensitivityMatrix,
    **kwargs: dict,
) -> float | complex:
    """Calculate the full Likelihood including noise and source terms.

    The noise term is calculated with :func:`noise_likelihood_term`.

    The source term is calcualted with :func:`residual_source_likelihood_term`.

    Args:
        data_res_arr: Data residual.
        psd: Sensitivity information.
        **kwargs: Keyword arguments to pass to :func:`inner_product`.

    Returns:
       Full Likelihood value.

    """
    if not isinstance(psd, SensitivityMatrix):
        # TODO: maybe adjust so it can take a list just like Sensitivity matrix
        psd = SensitivityMatrix(data_res_arr.f_arr, [psd], **kwargs)

    # remove key
    for key in "psd", "psd_args", "psd_kwargs":
        if key in kwargs:
            kwargs.pop(key)

    rslt = residual_source_likelihood_term(data_res_arr, psd=psd, **kwargs)

    nlt = noise_likelihood_term(psd)
    return nlt + rslt


def data_signal_source_likelihood_term(
    data_arr: DataResidualArray, sig_arr: DataResidualArray, **kwargs: dict
) -> float | complex:
    """Calculate the source term in the Likelihood for separate signal and data.

    The source term in the likelihood is given by,

    .. math::

        \\log{\\mathcal{L}}_\\text{src} = -\\frac{1}{2}\\left(\\langle \\vec{d} | \\vec{d}\\rangle + \\langle \\vec{h} | \\vec{h}\\rangle - 2\\langle \\vec{d} | \\vec{h}\\rangle \\right)\ \ .

    Args:
        data_arr: Data.
        sig_arr: Signal.
        **kwargs: Keyword arguments to pass to :func:`inner_product`.

    Returns:
        Source term Likelihood value.

    """
    kwargs["normalize"] = False
    d_h = inner_product(data_arr, sig_arr, **kwargs)
    h_h = inner_product(sig_arr, sig_arr, **kwargs)
    d_d = inner_product(data_arr, data_arr, **kwargs)
    return -1 / 2.0 * (d_d + h_h - 2 * d_h)


def data_signal_full_source_and_noise_likelihood(
    data_arr: DataResidualArray,
    sig_arr: DataResidualArray,
    psd: str | NoneType | np.ndarray | SensitivityMatrix,
    **kwargs: dict,
) -> float | complex:
    """Calculate the full Likelihood including noise and source terms.

    Here, the signal is treated separate from the data.

    The noise term is calculated with :func:`noise_likelihood_term`.

    The source term is calcualted with :func:`data_signal_source_likelihood_term`.

    Args:
        data_arr: Data.
        sig_arr: Signal.
        psd: Sensitivity information.
        **kwargs: Keyword arguments to pass to :func:`inner_product`.

    Returns:
       Full Likelihood value.

    """
    if not isinstance(psd, SensitivityMatrix):
        # TODO: maybe adjust so it can take a list just like Sensitivity matrix
        psd = SensitivityMatrix(data_arr.f_arr, [psd], **kwargs)

    # remove key
    for key in "psd", "psd_args", "psd_kwargs":
        if key in kwargs:
            kwargs.pop(key)

    rslt = data_signal_source_likelihood_term(data_arr, sig_arr, psd=psd, **kwargs)

    nlt = noise_likelihood_term(psd)

    return nlt + rslt


def snr(
    sig1: np.ndarray | list | DataResidualArray,
    *args: Any,
    data: Optional[np.ndarray | list | DataResidualArray] = None,
    **kwargs: Any,
) -> float:
    """Compute the snr between two signals weighted by a psd.

    The signal-to-noise ratio of a signal is :math:`\\sqrt{\\langle a|a\\rangle}`.

    This will be the optimal SNR if ``data==None``. If a data array is given, it will be the observed
    SNR: :math:`\\langle d|a\\rangle/\\sqrt{\\langle a|a\\rangle}`.

    **GPU Capability**: Pass CuPy arrays rather than NumPy arrays.

    Args:
        sig1: Signal to use as the templatefor the SNR.
            Can be time-domain or frequency-domain.
            Must be 1D ``np.ndarray``, list of 1D ``np.ndarray``s, or 2D ``np.ndarray``
            across channels with shape ``(nchannels, data length)``.
        *args: Arguments to pass to :func:`inner_product`.
        data: Data becomes the ``sig2`` argument to :func:`inner_product`.
        **kwargs: Keyword arguments to pass to :func:`inner_product`.

    Returns:
        Optimal or detected SNR value (depending on ``data`` kwarg).

    """

    # get optimal SNR
    opt_snr = np.sqrt(inner_product(sig1, sig1, *args, **kwargs).real)
    if data is None:
        return opt_snr

    else:
        # if inputed data, calculate detected SNR
        det_snr = inner_product(sig1, data, *args, **kwargs).real / opt_snr
        return det_snr


def h_var_p_eps(
    step: float,
    waveform_model: callable,
    params: ArrayLike,
    index: int,
    parameter_transforms: Optional[TransformContainer] = None,
    waveform_args: Optional[tuple] = (),
    waveform_kwargs: Optional[dict] = {},
) -> np.ndarray:  # TODO: check this
    """Calculate the waveform with a perturbation step of the variable V[i]

    Args:
        waveform_model: Callable function to the waveform generator with signature ``(*params, **waveform_kwargs)``.
        params: Source parameters that are over derivatives (not in fill dict of parameter transforms)
        step: Absolute step size for variable of interest.
        index: Index to parameter of interest.
        parameter_transforms: `TransformContainer <https://mikekatz04.github.io/Eryn/html/user/utils.html#eryn.utils.TransformContainer>`_ object to transform from the derivative parameter basis
            to the waveform parameter basis. This class can also fill in fixed parameters where the derivatives are not being taken.
        waveform_args: args (beyond parameters) for the waveform generator.
        waveform_kwargs: kwargs for the waveform generation.

    Returns:
        Perturbation to the waveform in the given parameter. Will always be 2D array with shape ``(num channels, data length)``

    """
    params_p_eps = params.copy()
    params_p_eps[index] += step

    if parameter_transforms is not None:
        # transform
        params_p_eps = parameter_transforms.transform_base_parameters(params_p_eps)

    args_in = tuple(params_p_eps) + tuple(waveform_args)
    dh = waveform_model(*args_in, **waveform_kwargs)

    # adjust output based on waveform model output
    # needs to be 2D array
    if (isinstance(dh, np.ndarray) or isinstance(dh, cp.ndarray)) and dh.ndim == 1:
        xp = get_array_module(dh)
        dh = xp.atleast_2d(dh)

    elif isinstance(dh, list):
        xp = get_array_module(dh[0])
        dh = xp.asarray(dh)

    return dh


def dh_dlambda(
    eps: float,
    *args: tuple,
    more_accurate: Optional[bool] = True,
    **kwargs: dict,
) -> np.ndarray:
    """Derivative of the waveform

    Calculate the derivative of the waveform with precision of order (step^4)
    with respect to the variable V in the i direction.

    Args:
        eps: Absolute **derivative** step size for variable of interest.
        *args: Arguments passed to :func:`h_var_p_eps`.
        more_accurate: If ``True``, run a more accurate derivate requiring 2x more waveform generations.
        **kwargs: Keyword arguments passed to :func:`h_var_p_eps`.

    Returns:
        Numerical derivative of the waveform with respect to a varibale of interest. Will be 2D array
        with shape ``(num channels, data length)``.

    """
    if more_accurate:
        # Derivative of the Waveform
        # up
        h_I_up_2eps = h_var_p_eps(2 * eps, *args, **kwargs)
        h_I_up_eps = h_var_p_eps(eps, *args, **kwargs)
        # down
        h_I_down_2eps = h_var_p_eps(-2 * eps, *args, **kwargs)
        h_I_down_eps = h_var_p_eps(-eps, *args, **kwargs)

        # make sure they are all the same length
        ind_max = np.min(
            [
                h_I_up_2eps.shape[1],
                h_I_up_eps.shape[1],
                h_I_down_2eps.shape[1],
                h_I_down_eps.shape[1],
            ]
        )

        # error scales as eps^4
        dh_I = (
            -h_I_up_2eps[:, :ind_max]
            + h_I_down_2eps[:, :ind_max]
            + 8 * (h_I_up_eps[:, :ind_max] - h_I_down_eps[:, :ind_max])
        ) / (12 * eps)

    else:
        # Derivative of the Waveform
        # up
        h_I_up_eps = h_var_p_eps(eps, *args, **kwargs)
        # down
        h_I_down_eps = h_var_p_eps(-eps, *args, **kwargs)

        ind_max = np.min([h_I_up_eps.shape[1], h_I_down_eps.shape[1]])

        # TODO: check what error scales as.
        dh_I = (h_I_up_eps[:, :ind_max] - h_I_down_eps[:, :ind_max]) / (2 * eps)
        # Time thta it takes for one variable: approx 5 minutes

    return dh_I


def info_matrix(
    eps: float | np.ndarray,
    waveform_model: callable,
    params: ArrayLike,
    deriv_inds: Optional[ArrayLike] = None,
    inner_product_kwargs: Optional[dict] = {},
    return_derivs: Optional[bool] = False,
    **kwargs: dict,
) -> np.ndarray | Tuple[np.ndarray, list]:
    """Calculate Information Matrix.

    This calculates the information matrix for a given waveform model at a given set of parameters.
    The inverse of the information matrix gives the covariance matrix.

    This is also referred to as the Fisher information matrix, but @MichaelKatz has chosen to leave out the name because of `this <https://www.newstatesman.com/long-reads/2020/07/ra-fisher-and-science-hatred>`_.

    The info matrix is given by:

    .. math::

        M_{ij} = \\langle h_i | h_j \\rangle \\text{ with } h_i = \\frac{\\partial h}{\\partial \\lambda_i}.

    Args:
        eps: Absolute **derivative** step size for variable of interest. Can be provided as a ``float`` value that applies
            to all variables or an array, one for each parameter being evaluated in the information matrix.
        waveform_model: Callable function to the waveform generator with signature ``(*params, **waveform_kwargs)``.
        params: Source parameters.
        deriv_inds: Subset of parameters of interest for which to calculate the information matrix, by index.
            If ``None``, it will be ``np.arange(len(params))``.
        inner_product_kwargs: Keyword arguments for the inner product function.
        return_derivs: If ``True``, also returns computed numerical derivatives.
        **kwargs: Keyword arguments passed to :func:`dh_dlambda`.

    Returns:
        If ``not return_derivs``, this will be the information matrix as a numpy array. If ``return_derivs is True``,
        it will be a tuple with the first entry as the information matrix and the second entry as the partial derivatives.

    """

    # setup initial information
    num_params = len(params)

    if deriv_inds is None:
        deriv_inds = np.arange(num_params)

    num_info_params = len(deriv_inds)

    if isinstance(eps, float):
        eps = np.full_like(params, eps)

    # collect derivatives over the variables of interest
    dh = []
    for i, eps_i in zip(deriv_inds, eps):
        # derivative up
        temp = dh_dlambda(eps_i, waveform_model, params, i, **kwargs)
        dh.append(temp)

    # calculate the components of the symmetric matrix
    info = np.zeros((num_info_params, num_info_params))
    for i in range(num_info_params):
        for j in range(i, num_info_params):
            info[i][j] = inner_product(
                [dh[i][k] for k in range(len(dh[i]))],
                [dh[j][k] for k in range(len(dh[i]))],
                **inner_product_kwargs,
            )
            info[j][i] = info[i][j]

    if return_derivs:
        return info, dh
    else:
        return info


def covariance(
    *info_mat_args: tuple,
    info_mat: Optional[np.ndarray] = None,
    diagonalize: Optional[bool] = False,
    return_info_mat: Optional[bool] = False,
    precision: Optional[bool] = False,
    **info_mat_kwargs: dict,
) -> np.ndarray | list:
    """Calculate covariance matrix for a set of EMRI parameters, computing the information matrix if not supplied.

    Args:
        *info_mat_args: Set of arguments to pass to :func:`info_matrix`. Not required if ``info_mat`` is not ``None``.
        info_mat: Pre-computed information matrix. If supplied, this matrix will be inverted.
        diagonalize: If ``True``, diagonalizes the covariance matrix.
        return_info_mat: If ``True``, also returns the computed information matrix.
        precision: If ``True``, uses 500-dps precision to compute the information matrix inverse (requires `mpmath <https://mpmath.org>`_). This is typically a good idea as the information matrix can be highly ill-conditioned.
        **info_mat_kwargs: Keyword arguments to pass to :func:`info_matrix`.

    Returns:
        Covariance matrix. If ``return_info_mat is True``. A list will be returned with the covariance as the first
        entry and the information matrix as the second entry. If ``return_derivs is True`` (keyword argument to :func:`info_matrix`),
        then another entry will be added to the list for the derivatives.

    """

    if info_mat is None:
        info_mat = info_matrix(*info_mat_args, **info_mat_kwargs)

    # parse output properly
    if "return_derivs" in info_mat_kwargs and info_mat_kwargs["return_derivs"]:
        return_derivs = True
        info_mat, dh = info_mat

    else:
        return_derivs = False

    # attempt to import and setup mpmath if precision required
    if precision:
        try:
            import mpmath as mp

            mp.mp.dps = 500
        except ModuleNotFoundError:
            print("mpmath module not installed. Defaulting to low precision...")
            precision = False

    if precision:
        hp_info_mat = mp.matrix(info_mat.tolist())
        U, S, V = mp.svd_r(hp_info_mat)  # singular value decomposition
        temp = mp.diag([val ** (-1) for val in S])  # get S**-1
        temp2 = V.T * temp * U.T  # construct pseudo-inverse
        cov = np.array(temp2.tolist(), dtype=np.float64)
    else:
        cov = np.linalg.pinv(info_mat)

    if diagonalize:
        # get eigeninformation
        eig_vals, eig_vecs = get_eigeninfo(cov, high_precision=precision)

        # diagonal cov now
        cov = np.dot(np.dot(eig_vecs.T, cov), eig_vecs)

    # just requesting covariance
    if True not in [return_info_mat, return_derivs]:
        return cov

    # if desiring more information, create a list capable of variable size
    returns = [cov]

    # add information matrix
    if return_info_mat:
        returns.append(info_mat)

    # add derivatives
    if return_derivs:
        returns.append(dh)

    return returns


def plot_covariance_corner(
    params: np.ndarray,
    cov: np.ndarray,
    nsamp: Optional[int] = 25000,
    fig: Optional[plt.Figure] = None,
    **kwargs: dict,
) -> plt.Figure:
    """Construct a corner plot for a given covariance matrix.

    The `corner <https://corner.readthedocs.io/en/latest/>`_ module is required for this.

    Args:
        params: The set of parameters used for the event (the mean vector of the covariance matrix).
        cov: Covariance matrix from which to construct the corner plot.
        nsamp: Number of samples to draw from the multivariate distribution.
        fig: Matplotlib :class:`plt.Figure` object. Use this if passing an existing corner plot figure.
        **kwargs: Keyword arguments for the corner plot - see the module documentation for more info.

    Returns:
        The corner plot figure.

    """

    # TODO: add capability for ChainConsumer?
    try:
        import corner
    except ModuleNotFoundError:
        raise ValueError(
            "Attempting to plot using the corner module, but it is not installed."
        )

    # generate fake samples from the covariance distribution
    samp = np.random.multivariate_normal(params, cov, size=nsamp)

    # make corner plot
    fig = corner.corner(samp, **kwargs)
    return fig


def plot_covariance_contour(
    params: np.ndarray,
    cov: np.ndarray,
    horizontal_index: int,
    vertical_index: int,
    nsamp: Optional[int] = 25000,
    ax: Optional[plt.Axes] = None,
    **kwargs: dict,
) -> plt.Axes | Tuple[plt.Figure, plt.Axes]:
    """Construct a contour plot for a given covariance matrix on a single axis object.

    The `corner <https://corner.readthedocs.io/en/latest/>`_ module is required for this.

    Args:
        params: The set of parameters used for the event (the mean vector of the covariance matrix).
        cov: Covariance matrix from which to construct the corner plot.
        horizontal_index: Parameter index to plot along the horizontal axis of the contour plot.
        vertical_index: Parameter index to plot along the vertical axis of the contour plot.
        nsamp: Number of samples to draw from the multivariate distribution.
        fig: Matplotlib :class:`plt.Figure` object. Use this if passing an existing corner plot figure.
        **kwargs: Keyword arguments for the corner plot - see the module documentation for more info.

    Returns:
        If ``ax`` is provided, the return will be that ax object. If it is not provided, then a
        Matplotlib Figure and Axes obejct is created and returned as a tuple: ``(plt.Figure, plt.Axes)``.

    """

    # TODO: add capability for ChainConsumer?
    try:
        import corner
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "Attempting to plot using the corner module, but it is not installed."
        )

    if ax is None:
        fig, ax = plt.subplots(1, 1)
    else:
        fig = None

    # generate fake samples from the covariance distribution
    samp = np.random.multivariate_normal(params, cov, size=nsamp)

    x = samp[:, horizontal_index]
    y = samp[:, vertical_index]

    # make corner plot
    corner.hist2d(x, y, ax=ax, **kwargs)

    if fig is None:
        return ax
    else:
        return (fig, ax)


def get_eigeninfo(
    arr: np.ndarray, high_precision: Optional[bool] = False
) -> Tuple[np.ndarray, np.ndarray]:
    """Performs eigenvalue decomposition and returns the eigenvalues and right-eigenvectors for the supplied fisher/covariance matrix.

    Args:
        arr: Input matrix for which to perform eigenvalue decomposition.
        high_precision: If ``True``, use 500-dps precision to ensure accurate eigenvalue decomposition
            (requires `mpmath <https://mpmath.org>`_ to be installed). Defaults to False.

    Returns:
        Tuple containing Eigenvalues and right-Eigenvectors for the supplied array, constructed such that evects[:,k] corresponds to evals[k].


    """

    if high_precision:
        try:
            import mpmath as mp

            mp.mp.dps = 500
        except ModuleNotFoundError:
            print("mpmath is not installed - using low-precision eigen decomposition.")
            high_precision = False

    if high_precision:
        hp_arr = mp.matrix(arr.tolist())
        # get eigenvectors
        E, EL, ER = mp.eig(hp_arr, left=True, right=True)

        # convert back
        evals = np.array(E, dtype=np.float64)
        evects = np.array(ER.tolist(), dtype=np.float64)

        return evals, evects

    else:
        evals, evects = np.linalg.eig(arr)
        return evals, evects


def cutler_vallisneri_bias(
    waveform_model_true: callable,
    waveform_model_approx: callable,
    params: np.ndarray,
    eps: float | np.ndarray,
    input_diagnostics: Optional[dict] = None,
    info_mat: Optional[np.ndarray] = None,
    deriv_inds: Optional[ArrayLike] = None,
    return_derivs: Optional[bool] = False,
    return_cov: Optional[bool] = False,
    parameter_transforms: Optional[TransformContainer] = None,
    waveform_true_args: Optional[tuple] = (),
    waveform_true_kwargs: Optional[dict] = {},
    waveform_approx_args: Optional[tuple] = (),
    waveform_approx_kwargs: Optional[dict] = {},
    inner_product_kwargs: Optional[dict] = {},
) -> list:
    """Calculate the Cutler-Vallisneri bias.

    # TODO: add basic math

    Args:
        waveform_model_true: Callable function to the **true** waveform generator with signature ``(*params, **waveform_kwargs)``.
        waveform_model_approx: Callable function to the **approximate** waveform generator with signature ``(*params, **waveform_kwargs)``.
        params: Source parameters.
        eps: Absolute **derivative** step size. See :func:`info_matrix`.
        input_diagnostics: Dictionary including the diagnostic information if it is precomputed. Dictionary must include
            keys ``"cov"`` (covariance matrix, output of :func:`covariance`), ``"h_true"`` (the **true** waveform),
            and ``"dh"`` (derivatives of the waveforms, list of outputs from :func:`dh_dlambda`).
        info_mat: Pre-computed information matrix. If supplied, this matrix will be inverted to find the covariance.
        deriv_inds: Subset of parameters of interest. See :func:`info_matrix`.
        return_derivs: If ``True``, also returns computed numerical derivatives.
        return_cov: If ``True``, also returns computed covariance matrix.
        parameter_transforms: `TransformContainer <https://mikekatz04.github.io/Eryn/html/user/utils.html#eryn.utils.TransformContainer>`_ object. See :func:`info_matrix`.
        waveform_true_args: Arguments for the **true** waveform generator.
        waveform_true_kwargs: Keyword arguments for the **true** waveform generator.
        waveform_approx_args: Arguments for the **approximate** waveform generator.
        waveform_approx_kwargs: Keyword arguments for the **approximate** waveform generator.
        inner_product_kwargs: Keyword arguments for the inner product function.

    Returns:
        List of return information. By default, it is ``[systematic error, bias]``.
        If ``return_derivs`` or ``return_cov`` are ``True``, they will be added to the list with derivs added before covs.

    """

    if deriv_inds is None:
        deriv_inds = np.arange(len(params))

    if info_mat is not None and input_diagnostics is not None:
        warnings.warn(
            "Provided info_mat and input_diagnostics kwargs. Ignoring info_mat."
        )

    # adjust parameters to waveform basis
    params_in = parameter_transforms.transform_base_parameters(params.copy())

    if input_diagnostics is None:
        # get true waveform
        h_true = waveform_model_true(
            *(tuple(params_in) + tuple(waveform_true_args)), **waveform_true_kwargs
        )

        # get covariance info and waveform derivatives
        cov, dh = covariance(
            eps,
            waveform_model_true,
            params,
            return_derivs=True,
            deriv_inds=deriv_inds,
            info_mat=info_mat,
            parameter_transforms=parameter_transforms,
            waveform_args=waveform_true_args,
            waveform_kwargs=waveform_true_kwargs,
            inner_product_kwargs=inner_product_kwargs,
        )

    else:
        # pre-computed info
        cov = input_diagnostics["cov"]
        h_true = input_diagnostics["h_true"]
        dh = input_diagnostics["dh"]

    # get approximate waveform
    h_approx = waveform_model_approx(
        *(tuple(params_in) + tuple(waveform_approx_args)), **waveform_approx_kwargs
    )

    # adjust/check waveform outputs
    if isinstance(h_true, np.ndarray) and h_true.ndim == 1:
        h_true = [h_true]
    elif isinstance(h_true, np.ndarray) and h_true.ndim == 2:
        h_true = list(h_true)

    if isinstance(h_approx, np.ndarray) and h_approx.ndim == 1:
        h_approx = [h_approx]
    elif isinstance(h_approx, np.ndarray) and h_approx.ndim == 2:
        h_approx = list(h_approx)

    assert len(h_approx) == len(h_true)
    assert np.all(
        np.asarray([len(h_approx[i]) == len(h_true[i]) for i in range(len(h_true))])
    )

    # difference in the waveforms
    diff = [h_true[i] - h_approx[i] for i in range(len(h_approx))]

    # systematic err
    syst_vec = np.array(
        [
            inner_product(
                dh[k],
                diff,
                **inner_product_kwargs,
            )
            for k in range(len(deriv_inds))
        ]
    )

    bias = np.dot(cov, syst_vec)

    # return list
    returns = [syst_vec, bias]

    # add anything requested
    if return_cov:
        returns.append(cov)

    if return_derivs:
        returns.append(dh)

    return returns


def scale_to_snr(
    target_snr: float,
    sig: np.ndarray | list,
    *snr_args: tuple,
    return_orig_snr: Optional[bool] = False,
    **snr_kwargs: dict,
) -> np.ndarray | list | Tuple[np.ndarray | list, float]:
    """Calculate the SNR and scale a signal.

    Args:
        target_snr: Desired SNR value for the injected signal.
        sig: Signal to adjust. A copy will be made.
            Can be time-domain or frequency-domain.
            Must be 1D ``np.ndarray``, list of 1D ``np.ndarray``s, or 2D ``np.ndarray``
            across channels with shape ``(nchannels, data length)``.
        *snr_args: Arguments to pass to :func:`snr`.
        return_orig_snr: If ``True``, return the original SNR in addition to the adjusted data.
        **snr_kwargs: Keyword arguments to pass to :func:`snr`.

    Returns:
        Returns the copied input signal adjusted to the target SNR. If ``return_orig_snr is True``, the original
        SNR is added as the second entry of a tuple with the adjusted signal (as the first entry in the tuple).

    """
    # get the snr and adjustment factor
    snr_out = snr(sig, *snr_args, **snr_kwargs)
    factor = target_snr / snr_out

    # any changes back to the original signal type
    back_single = False
    back_2d_array = False

    if isinstance(sig, list) is False and sig.ndim == 1:
        sig = [sig]
        back_single = True

    elif isinstance(sig, list) is False and sig.ndim == 2:
        sig = list(sig)
        back_2d_array = True

    # adjust
    out = [sig_i * factor for sig_i in sig]

    # adjust type back to the input type
    if back_2d_array:
        out = np.asarray(out)

    if back_single:
        out = out[0]

    if return_orig_snr:
        return (out, snr_out)

    return out



import os
from abc import ABC, abstractmethod
from typing import Any, List, Tuple, Optional
from dataclasses import dataclass
import requests

import numpy as np
import h5py
from scipy import interpolate

from lisatools.cutils.detector import pycppDetector

SC = [1, 2, 3]
LINKS = [12, 23, 31, 13, 32, 21]

LINEAR_INTERP_TIMESTEP = 600.00  # sec (0.25 hr)


class Orbits(ABC):
    """LISA Orbit Base Class

    Args:
        filename: File name. File should be in the style of LISAOrbits

    """

    def __init__(self, filename: str) -> None:
        self.filename = filename
        self._setup()
        self.configured = False

    @property
    def LINKS(self) -> List[int]:
        """Link order."""
        return LINKS

    @property
    def SC(self) -> List[int]:
        """Spacecraft order."""
        return SC

    @property
    def link_space_craft_r(self) -> List[int]:
        """Receiver (first) spacecraft"""
        return [int(str(link_i)[0]) for link_i in self.LINKS]

    @property
    def link_space_craft_e(self) -> List[int]:
        """Sender (second) spacecraft"""
        return [int(str(link_i)[1]) for link_i in self.LINKS]

    def _setup(self) -> None:
        """Read in orbital data from file and store."""
        with self.open() as f:
            for key in f.attrs.keys():
                setattr(self, key + "_base", f.attrs[key])

    @property
    def filename(self) -> str:
        """Orbit file name."""
        return self._filename

    @filename.setter
    def filename(self, filename: str) -> None:
        """Set file name."""

        assert isinstance(filename, str)

        # get path
        path_to_this_file = __file__.split("detector.py")[0]

        # make sure orbit_files directory exists in the right place
        if not os.path.exists(path_to_this_file + "orbit_files/"):
            os.mkdir(path_to_this_file + "orbit_files/")
        path_to_this_file = path_to_this_file + "orbit_files/"

        if not os.path.exists(path_to_this_file + filename):
            # download files from github if they are not there
            github_file = f"https://github.com/mikekatz04/LISAanalysistools/raw/main/lisatools/orbit_files/{filename}"
            r = requests.get(github_file)

            # if not success
            if r.status_code != 200:
                raise ValueError(
                    f"Cannot find {filename} within default files located at github.com/mikekatz04/LISAanalysistools/lisatools/orbit_files/."
                )

            # write the contents to a local file
            with open(path_to_this_file + filename, "wb") as f:
                f.write(r.content)

        # store
        self._filename = path_to_this_file + filename

    def open(self) -> h5py.File:
        """Opens the h5 file in the proper mode.

        Returns:
            H5 file object: Opened file.

        Raises:
            RuntimeError: If backend is opened for writing when it is read-only.

        """
        f = h5py.File(self.filename, "r")
        return f

    @property
    def t_base(self) -> np.ndarray:
        """Time array from file."""
        with self.open() as f:
            t_base = np.arange(self.size_base) * self.dt_base
        return t_base

    @property
    def ltt_base(self) -> np.ndarray:
        """Light travel times along links from file."""
        with self.open() as f:
            ltt = f["tcb"]["ltt"][:]
        return ltt

    @property
    def n_base(self) -> np.ndarray:
        """Normal unit vectors towards receiver along links from file."""
        with self.open() as f:
            n = f["tcb"]["n"][:]
        return n

    @property
    def x_base(self) -> np.ndarray:
        """Spacecraft position from file."""
        with self.open() as f:
            x = f["tcb"]["x"][:]
        return x

    @property
    def v_base(self) -> np.ndarray:
        """Spacecraft velocities from file."""
        with self.open() as f:
            v = f["tcb"]["v"][:]
        return v

    @property
    def t(self) -> np.ndarray:
        """Configured time array."""
        self._check_configured()
        return self._t

    @t.setter
    def t(self, t: np.ndarray):
        """Set configured time array."""
        assert isinstance(t, np.ndarray) and t.ndim == 1
        self._t = t

    @property
    def ltt(self) -> np.ndarray:
        """Light travel time."""
        self._check_configured()
        return self._ltt

    @ltt.setter
    def ltt(self, ltt: np.ndarray) -> np.ndarray:
        """Set light travel time."""
        assert ltt.shape[0] == len(self.t)

    @property
    def n(self) -> np.ndarray:
        """Normal vectors along links."""
        self._check_configured()
        return self._n

    @n.setter
    def n(self, n: np.ndarray) -> np.ndarray:
        """Set Normal vectors along links."""
        return self._n

    @property
    def x(self) -> np.ndarray:
        """Spacecraft positions."""
        self._check_configured()
        return self._x

    @x.setter
    def x(self, x: np.ndarray) -> np.ndarray:
        """Set Spacecraft positions."""
        return self._x

    @property
    def v(self) -> np.ndarray:
        """Spacecraft velocities."""
        self._check_configured()
        return self._v

    @v.setter
    def v(self, v: np.ndarray) -> np.ndarray:
        """Set Spacecraft velocities."""
        return self._v

    def configure(
        self,
        t_arr: Optional[np.ndarray] = None,
        dt: Optional[float] = None,
        linear_interp_setup: Optional[bool] = False,
    ) -> None:
        """Configure the orbits to match the signal response generator time basis.

        The base orbits will be scaled up or down as needed using Cubic Spline interpolation.
        The higherarchy of consideration to each keyword argument if multiple are given:
        ``linear_interp_setup``, ``t_arr``, ``dt``.

        If nothing is provided, the base points are used.

        Args:
            t_arr: New time array.
            dt: New time step. Will take the time duration to be that of the input data.
            linear_interp_setup: If ``True``, it will create a dense grid designed for linear interpolation with a constant time step.

        """

        x_orig = self.t_base

        if linear_interp_setup:
            make_cpp = True
            dt = LINEAR_INTERP_TIMESTEP
            Tobs = self.t_base[-1]
            Nobs = int(Tobs / dt)
            t_arr = np.arange(Nobs) * dt
            if t_arr[-1] < self.t_base[-1]:
                t_arr = np.concatenate([t_arr, self.t_base[-1:]])
        elif t_arr is not None:
            assert np.all(t_arr >= self.t_base[0]) and np.all(t_arr <= self.t_base[-1])

        elif dt is not None:
            make_cpp = True
            Tobs = self.t_base[-1]
            Nobs = int(Tobs / dt)
            t_arr = np.arange(Nobs) * dt
            if t_arr[-1] < self.t_base[-1]:
                t_arr = np.concatenate([t_arr, self.t_base[-1:]])

        else:
            make_cpp = False
            t_arr = self.t_base

        x_new = t_arr.copy()
        self.t = t_arr.copy()

        for which in ["ltt", "x", "n", "v"]:
            arr = getattr(self, which + "_base")
            arr_tmp = arr.reshape(self.size_base, -1)
            arr_out_tmp = np.zeros((len(x_new), arr_tmp.shape[-1]))
            for i in range(arr_tmp.shape[-1]):
                arr_out_tmp[:, i] = interpolate.CubicSpline(x_orig, arr_tmp[:, i])(
                    x_new
                )
            arr_out = arr_out_tmp.reshape((len(x_new),) + arr.shape[1:])
            setattr(self, "_" + which, arr_out)

        self.configured = True

        lsr = np.asarray(self.link_space_craft_r).copy().astype(np.int32)
        lse = np.asarray(self.link_space_craft_e).copy().astype(np.int32)
        ll = np.asarray(self.LINKS).copy().astype(np.int32)

        if make_cpp:
            self.pycppdetector_args = (
                dt,
                len(self.t),
                self.n.flatten().copy(),
                self.ltt.flatten().copy(),
                self.x.flatten().copy(),
                ll,
                lsr,
                lse,
            )
            self.dt = dt
        else:
            self.pycppdetector_args = None
            self.dt = dt

    @property
    def dt(self) -> float:
        """new time step if it exists"""
        if self._dt is None:
            raise ValueError("dt not available for t_arr only.")
        return self._dt

    @dt.setter
    def dt(self, dt: float) -> None:
        self._dt = dt

    @property
    def pycppdetector(self) -> pycppDetector:
        """C++ class"""
        if self._pycppdetector_args is None:
            raise ValueError(
                "Asking for c++ class. Need to set linear_interp_setup = True when configuring."
            )
        self._pycppdetector = pycppDetector(*self._pycppdetector_args)
        return self._pycppdetector

    @property
    def pycppdetector_args(self) -> tuple:
        """args for the c++ class."""
        return self._pycppdetector_args

    @pycppdetector_args.setter
    def pycppdetector_args(self, pycppdetector_args: tuple) -> None:
        self._pycppdetector_args = pycppdetector_args

    @property
    def size(self) -> int:
        """Number of time points."""
        self._check_configured()
        return len(self.t)

    def _check_configured(self) -> None:
        if not self.configured:
            raise ValueError(
                "Cannot request property. Need to use configure() method first."
            )

    def get_light_travel_times(
        self, t: float | np.ndarray, link: int
    ) -> float | np.ndarray:
        """Compute light travel time as a function of time.

        Computes with the c++ backend.

        Args:
            t: Time array in seconds.
            link: which link. Must be ``in self.LINKS``.

        Returns:
            Light travel times.

        """
        return self.pycppdetector.get_light_travel_time(t, link)

    def get_normal_unit_vec(self, t: float | np.ndarray, link: int) -> np.ndarray:
        """Compute link normal vector as a function of time.

        Computes with the c++ backend.

        Args:
            t: Time array in seconds.
            link: which link. Must be ``in self.LINKS``.

        Returns:
            Link normal vectors.

        """
        return self.pycppdetector.get_normal_unit_vec(t, link)

    def get_pos(self, t: float | np.ndarray, sc: int) -> np.ndarray:
        """Compute spacecraft position as a function of time.

        Computes with the c++ backend.

        Args:
            t: Time array in seconds.
            sc: which spacecraft. Must be ``in self.SC``.

        Returns:
            Spacecraft positions.

        """
        return self.pycppdetector.get_pos(t, sc)

    @property
    def ptr(self) -> int:
        """pointer to c++ class"""
        return self.pycppdetector.ptr


class EqualArmlengthOrbits(Orbits):
    """Equal Armlength Orbits

    Orbit file: equalarmlength-orbits.h5

    """

    def __init__(self):
        super().__init__("equalarmlength-orbits.h5")


class ESAOrbits(Orbits):
    """ESA Orbits

    Orbit file: esa-trailing-orbits.h5

    """

    def __init__(self):
        # TODO: fix this up
        super().__init__("esa-trailing-orbits.h5")


class DefaultOrbits(EqualArmlengthOrbits):
    """Set default orbit class to Equal Armlength orbits for now."""

    pass


@dataclass
class LISAModelSettings:
    """Required LISA model settings:

    Args:
        Soms_d: OMS displacement noise.
        Sa_a: Acceleration noise.
        orbits: Orbital information.
        name: Name of model.

    """

    Soms_d: float
    Sa_a: float
    orbits: Orbits
    name: str


class LISAModel(LISAModelSettings, ABC):
    """Model for the LISA Constellation

    This includes sensitivity information computed in
    :module:`lisatools.sensitivity` and orbital information
    contained in an :class:`Orbits` class object.

    This class is used to house high-level methods useful
    to various needed computations.

    """

    def __str__(self) -> str:
        out = "LISA Constellation Configurations Settings:\n"
        for key, item in self.__dict__.items():
            out += f"{key}: {item}\n"
        return out


# defaults
scirdv1 = LISAModel((15.0e-12) ** 2, (3.0e-15) ** 2, DefaultOrbits(), "scirdv1")
proposal = LISAModel((10.0e-12) ** 2, (3.0e-15) ** 2, DefaultOrbits(), "proposal")
mrdv1 = LISAModel((10.0e-12) ** 2, (2.4e-15) ** 2, DefaultOrbits(), "mrdv1")
sangria = LISAModel((10.0e-12) ** 2, (2.4e-15) ** 2, DefaultOrbits(), "sangria")

__stock_list_models__ = [scirdv1, proposal, mrdv1, sangria]
__stock_list_models_name__ = [tmp.name for tmp in __stock_list_models__]


def get_available_default_lisa_models() -> List[LISAModel]:
    """Get list of default LISA models

    Returns:
        List of LISA models.

    """
    return __stock_list_models__


def get_default_lisa_model_from_str(model: str) -> LISAModel:
    """Return a LISA model from a ``str`` input.

    Args:
        model: Model indicated with a ``str``.

    Returns:
        LISA model associated to that ``str``.

    """
    if model not in __stock_list_models_name__:
        raise ValueError(
            "Requested string model is not available. See lisatools.detector documentation."
        )
    return globals()[model]


def check_lisa_model(model: Any) -> LISAModel:
    """Check input LISA model.

    Args:
        model: LISA model to check.

    Returns:
        LISA Model checked. Adjusted from ``str`` if ``str`` input.

    """
    if isinstance(model, str):
        model = get_default_lisa_model_from_str(model)

    if not isinstance(model, LISAModel):
        raise ValueError("model argument not given correctly.")

    return model



# -*- coding: utf-8 -*-
import os.path
from tarfile import open as opentar

import gdown as gdd

os.makedirs("tutorial/data/", exist_ok=True)

# download MBH
if not os.path.exists("tutorial/data/mbh"):
    url = "https://drive.google.com/file/d/1Rr3pEF2BA2oRfs23FDIgMgoutxTq1qOR/view?usp=share_link"  # noqa: E501
    output_path = "tutorial/data/MBHDemo.tar"
    gdd.download(url, output_path, quiet=False, fuzzy=True)
    with opentar("tutorial/data/MBHDemo.tar", "r") as tar:
        tar.extractall(
            "tutorial/data/mbh"
        )  # specify which folder to extract to
    os.remove("tutorial/data/MBHDemo.tar")


# download UCB
if not os.path.exists("tutorial/data/ucb"):
    # 06 mo
    url = "https://drive.google.com/file/d/1z4M2cZnF5mPsS6ppNsHtceqXfR9Liz99/view?usp=share_link"  # noqa: E501
    output_path = "tutorial/data/UCBDemo_06.tar"
    gdd.download(url, output_path, quiet=False, fuzzy=True)
    with opentar("tutorial/data/UCBDemo_06.tar", "r") as tar:
        tar.extractall(
            "tutorial/data/ucb"
        )  # specify which folder to extract to
    os.remove("tutorial/data/UCBDemo_06.tar")

    # 03 mo
    url = "https://drive.google.com/file/d/1XGYUETwupZe8AiEDd4OMAIzA7kukDh5e/view?usp=share_link"  # noqa: E501
    output_path = "tutorial/data/UCBDemo_03.tar"
    gdd.download(url, output_path, quiet=False, fuzzy=True)
    with opentar("tutorial/data/UCBDemo_03.tar", "r") as tar:
        tar.extractall(
            "tutorial/data/ucb"
        )  # specify which folder to extract to
    os.remove("tutorial/data/UCBDemo_03.tar")



# -*- coding: utf-8 -*-

import numpy as np

from eryn.moves import MHMove

__all__ = ["GaussianMove"]


class SkyMove(MHMove):
    """A Metropolis step with a Gaussian proposal function.

    Args:


    Raises:
        ValueError: If the proposal dimensions are invalid or if any of any of
            the other arguments are inconsistent.

    """

    def __init__(self, ind_map=None, which="both", **kwargs):

        if ind_map is None:
            ind_map = dict(cosinc=6, lam=7, sinbeta=8, psi=9)

        elif isinstance(ind_map, dict) is False:
            raise ValueError("If providing the ind_map kwarg, it must be a dict.")

        if which not in ["both", "lat", "long"]:
            raise ValueError("which kwarg must be 'both', 'lat', or 'long'.")

        self.ind_map = ind_map
        self.which = which
        exec(f"self.transform = self.{which}_transform")
        super(SkyMove, self).__init__(**kwargs)

    def lat_transform(self, coords, random):
        """
        assumes sin beta
        assumes 2d array with all coords
        coords[]
        """
        temp = coords.copy()

        temp[:, self.ind_map["sinbeta"]] *= -1
        temp[:, self.ind_map["cosinc"]] *= -1
        temp[:, self.ind_map["psi"]] = np.pi - temp[:, self.ind_map["psi"]]

        return temp

    def long_transform(self, coords, random):
        """
        assumes sin beta
        assumes 2d array with all coords
        coords[]
        """
        temp = coords.copy()

        move_amount = random.randint(0, 4, size=coords.shape[0]) * np.pi / 2.0

        temp[:, self.ind_map["psi"]] += move_amount
        temp[:, self.ind_map["lam"]] += move_amount

        temp[:, self.ind_map["psi"]] %= np.pi
        temp[:, self.ind_map["lam"]] %= 2 * np.pi

        return temp

    def both_transform(self, coords, random):

        # if doing both does not assume it will cross plane, selects from 8 modes
        inds_lat_change = random.randint(0, 2, size=coords.shape[0]).astype(bool)
        coords[inds_lat_change] = self.lat_transform(coords[inds_lat_change], random)
        coords = self.long_transform(coords, random)
        return coords

    def get_proposal(self, branches_coords, random, branches_inds=None, **kwargs):
        """Get proposal from Gaussian distribution

        Args:
            branches_coords (dict): Keys are ``branch_names`` and values are
                np.ndarray[nwalkers, nleaves_max, ndim] representing
                coordinates for walkers.
            branches_inds (dict): Keys are ``branch_names`` and values are
                np.ndarray[nwalkers, nleaves_max] representing which
                leaves are currently being used.
            random (object): Current random state object.

        """

        q = {}
        for name, coords in zip(
            branches_coords.keys(), branches_coords.values()
        ):

            if branches_inds is None:
                inds = np.ones(coords.shape[:-1], dtype=bool)

            else:
                inds = branches_inds[name]

            ntemps, nwalkers, _, _ = coords.shape
            inds_here = np.where(inds == True)

            q[name] = coords.copy()
            new_coords = self.transform(coords[inds_here], random)
            q[name][inds_here] = new_coords.copy()

        return q, np.zeros((ntemps, nwalkers))



import numpy as np

from eryn.moves import Move

class PlaceHolder(Move):
    def __init__(self, *args, **kwargs):
        super(PlaceHolder, self).__init__(*args, **kwargs)

    def propose(self, model, state):
        accepted = np.zeros(state.log_prob.shape)
        try:
            self.temperature_control.swaps_accepted = np.zeros(len(self.temperature_control.betas) - 1, dtype=int)
        except AttributeError:
            pass

        return state, accepted


import numpy as np
cimport numpy as np
from libcpp.string cimport string
from libcpp cimport bool
from lisatools.utils.pointeradjust import wrapper
from libc.stdint cimport uintptr_t


cdef extern from "../include/Detector.hpp":
    cdef cppclass VecWrap "Vec":
        double x
        double y
        double z
        VecWrap(double x_, double y_, double z_) except+

    cdef cppclass OrbitsWrap "Orbits":
        OrbitsWrap(double dt_, int N_, double *n_arr_, double *L_arr_, double *x_arr_, int *links_, int *sc_r_, int *sc_e_) except+
        int get_window(double t) except+
        void get_normal_unit_vec_ptr(VecWrap *vec, double t, int link)
        int get_link_ind(int link) except+
        int get_sc_ind(int sc) except+
        double get_light_travel_time(double t, int link) except+
        VecWrap get_pos_ptr(VecWrap* out, double t, int sc) except+
        void dealloc();
        

cdef class pycppDetector:
    cdef OrbitsWrap *g
    cdef double dt
    cdef int N
    cdef size_t n_arr
    cdef size_t L_arr
    cdef size_t x_arr
    cdef size_t links
    cdef size_t sc_r
    cdef size_t sc_e

    def __cinit__(self, 
        *args, 
        **kwargs
    ):
        (
            dt,
            N, 
            n_arr,
            L_arr, 
            x_arr,
            links,
            sc_r, 
            sc_e
        ), tkwargs = wrapper(*args, **kwargs)

        self.dt = dt
        self.N = N 
        self.n_arr = n_arr
        self.L_arr = L_arr 
        self.x_arr = x_arr
        self.links = links
        self.sc_r = sc_r
        self.sc_e = sc_e

        cdef size_t n_arr_in = n_arr
        cdef size_t L_arr_in = L_arr
        cdef size_t x_arr_in = x_arr
        cdef size_t links_in = links
        cdef size_t sc_r_in = sc_r
        cdef size_t sc_e_in = sc_e
        
        self.g = new OrbitsWrap(
            dt,
            N,
            <double*> n_arr_in,
            <double*> L_arr_in, 
            <double*> x_arr_in, 
            <int*> links_in, 
            <int*> sc_r_in, 
            <int*> sc_e_in
        )

    def __dealloc__(self):
        self.g.dealloc()
        if self.g:
            del self.g

    def __reduce__(self):
        return (rebuild, (self.dt, self.N, self.n_arr, self.L_arr, self.x_arr, self.links, self.sc_r, self.sc_e,))

    def get_window(self, t: float) -> int:
        return self.g.get_window(t)

    def get_normal_unit_vec_single(self, t: float, link: int) -> np.ndarray:
        cdef VecWrap *out = new VecWrap(0.0, 0.0, 0.0)
        self.g.get_normal_unit_vec_ptr(out, t, link)

        return np.array([out.x, out.y, out.z])

    def get_normal_unit_vec_arr(self, t: np.ndarray, link: int) -> np.ndarray:
        output = np.zeros((len(t), 3), dtype=float)
        assert t.ndim == 1
        for i in range(len(t)):
            output[i] = self.get_normal_unit_vec_single(t[i], link)

        return output

    def get_normal_unit_vec(self, t: np.ndarray | float, link: int) -> np.ndarray:

        if isinstance(t, float):
            return self.get_normal_unit_vec_single(t, link)
        elif isinstance(t, np.ndarray):
            return self.get_normal_unit_vec_arr(t, link)

    def get_link_ind(self, link: int) -> int:
        return self.g.get_link_ind(link)

    def get_sc_ind(self, sc: int) -> int:
        return self.g.get_sc_ind(sc)

    def get_light_travel_time_single(self, t: float, link: int) -> float:
        return self.g.get_light_travel_time(t, link)

    def get_light_travel_time_arr(self, t: np.ndarray, link: int) -> np.ndarray:
        output = np.zeros((len(t),), dtype=float)
        assert t.ndim == 1
        for i in range(len(t)):
            output[i] = self.get_light_travel_time_single(t[i], link)

        return output

    def get_light_travel_time(self, t: np.ndarray | float, link: int) -> np.ndarray:

        if isinstance(t, float):
            print("t", t)
            return self.get_light_travel_time_single(t, link)
        elif isinstance(t, np.ndarray):
            return self.get_light_travel_time_arr(t, link)
    
    def get_pos_single(self, t: float, sc: int) -> np.ndarray:
  
        cdef VecWrap *out = new VecWrap(0.0, 0.0, 0.0)
        self.g.get_pos_ptr(out, t, sc)

        return np.array([out.x, out.y, out.z])

    def get_pos_arr(self, t: np.ndarray, sc: int) -> np.ndarray:
        output = np.zeros((len(t), 3), dtype=float)
        assert t.ndim == 1
        for i in range(len(t)):
            output[i] = self.get_pos_single(t[i], sc)

        return output

    def get_pos(self, t: np.ndarray | float, sc: int) -> np.ndarray:

        if isinstance(t, float):
            return self.get_pos_single(t, sc)
        elif isinstance(t, np.ndarray):
            return self.get_pos_arr(t, sc)

    @property
    def ptr(self) -> long:
        return <uintptr_t>self.g
    
        
def rebuild(dt,
    N, 
    n_arr,
    L_arr, 
    x_arr,
    links,
    sc_r, 
    sc_e
):
    c = pycppDetector(
        dt,
        N, 
        n_arr,
        L_arr, 
        x_arr,
        links,
        sc_r, 
        sc_e
    )
    return c


# -*- coding: utf-8 -*-

from copy import deepcopy
from inspect import Attribute
from multiprocessing.sharedctypes import Value
import numpy as np
import warnings
import time
from scipy.special import logsumexp

try:
    import cupy as xp

    gpu_available = True
except ModuleNotFoundError:
    import numpy as xp

    gpu_available = False

from eryn.state import State, BranchSupplimental
from eryn.moves import ReversibleJumpMove, MultipleTryMove
from eryn.prior import ProbDistContainer
from eryn.utils.utility import groups_from_inds

from gbgpu.utils.utility import get_N, get_fdot

from lisatools.sampling.moves.gbspecialstretch import GBSpecialStretchMove

__all__ = ["GBMutlipleTryRJ"]


def shuffle_along_axis(a, axis):
    idx = np.random.rand(*a.shape).argsort(axis=axis)
    return np.take_along_axis(a, idx, axis=axis)


def searchsorted2d_vec(a, b, xp=None, **kwargs):
    if xp is None:
        xp = np
    m, n = a.shape
    max_num = xp.maximum(a.max() - a.min(), b.max() - b.min()) + 1
    r = max_num * xp.arange(a.shape[0])[:, None]
    p = xp.searchsorted((a + r).ravel(), (b + r).ravel(), **kwargs).reshape(m, -1)
    return p - n * (xp.arange(m)[:, None])


class GBMutlipleTryRJ(MultipleTryMove, ReversibleJumpMove, GBSpecialStretchMove):
    """Generate Revesible-Jump proposals for GBs with multiple try

    Will use gpu if template generator uses GPU.

    Args:
        priors (object): :class:`PriorContainer` object that has ``logpdf``
            and ``rvs`` methods.

    """

    def __init__(
        self,
        gb_args,
        gb_kwargs,
        m_chirp_lims,
        *args,
        start_ind_limit=10,
        num_try=1,
        point_generator_func=None,
        fix_change=None,
        **kwargs
    ):
        self.point_generator_func = point_generator_func
        self.fixed_like_diff = 0
        self.time = 0
        self.name = "gbgroupstretch"
        self.start_ind_limit = start_ind_limit
        self.num_try = num_try

        GBSpecialStretchMove.__init__(self, *gb_args, **gb_kwargs)
        ReversibleJump.__init__(self, *args, **kwargs)
        MultipleTryMove.__init__(self, self.num_try, take_max_ll=False, xp=self.xp)

        # setup band edges for priors
        self.band_edges_fdot = self.xp.zeros_like(self.band_edges)
        # lower limit

        self.band_edges_fdot[:-1] = self.xp.asarray(
            get_fdot(self.band_edges[:-1], Mc=m_chirp_lims[0])
        )
        self.band_edges_fdot[1:] = self.xp.asarray(
            get_fdot(self.band_edges[1:], Mc=m_chirp_lims[1])
        )

        self.band_edges_gpu = self.xp.asarray(self.band_edges)
        self.fix_change = fix_change
        if self.fix_change not in [None, +1, -1]:
            raise ValueError("fix_change must be None, +1, or -1.")

    def special_generate_func(
        self,
        coords,
        nwalkers,
        current_priors=None,
        random=None,
        size: int = 1,
        fill=None,
        fill_inds=None,
        band_inds=None,
    ):
        """if self.search_samples is not None:
        # TODO: make replace=True ? in PE
        inds_drawn = self.xp.array([random.choice(
            self.search_inds, size=size, replace=False,
        ) for w in range(nwalkers)])
        generated_points = self.search_samples[inds_drawn].copy()  # .reshape(nwalkers, size, -1)
        # since this is in search only, we will pretend these are coming from the prior
        # so that they are accepted based on the Likelihood (and not the posterior)
        generate_factors = current_priors.logpdf(generated_points.reshape(nwalkers * size, -1)).reshape(nwalkers, size)
        """
        # st = time.perf_counter()
        if band_inds is None:
            raise ValueError("band_inds needs to be set")

        # elif
        if self.point_generator_func is not None:
            # st1 = time.perf_counter()
            generated_points = self.point_generator_func.rvs(size=size * nwalkers)
            """et1 = time.perf_counter()
            print("generate rvs:", et1 - st1)"""

            generated_points[self.xp.isnan(generated_points[:, 0]), 0] = 0.01

            # st2 = time.perf_counter()
            generate_factors = self.point_generator_func.logpdf(generated_points)
            """et2 = time.perf_counter()
            print("generate logpdf:", et2 - st2)"""

            starts = self.band_edges_gpu[band_inds]
            ends = self.band_edges_gpu[band_inds + 1]

            starts_fdot = self.band_edges_fdot[band_inds]
            ends_fdot = self.band_edges_fdot[band_inds + 1]

            # fill before getting logpdf
            generated_points = generated_points.reshape(nwalkers, size, -1)
            generate_factors = generate_factors.reshape(nwalkers, size)

            # map back per band
            generated_points[:, :, 1] = (
                generated_points[:, :, 1] * (ends[:, None] - starts[:, None])
                + starts[:, None]
            ) * 1e3
            generated_points[:, :, 2] = (
                generated_points[:, :, 2] * (ends_fdot[:, None] - starts_fdot[:, None])
                + starts_fdot[:, None]
            )

            if fill is not None or fill_inds is not None:
                if fill is None or fill_inds is None:
                    raise ValueError(
                        "If providing fill_inds or fill, must provide both."
                    )
                generated_points[fill_inds] = fill.copy()

            generated_points = generated_points.reshape(nwalkers, size, -1)

            # logpdf contribution from original distribution is zero = log(1/1)
            # THIS HAS BEEN REMOVED TO SIMULATE A PRIOR HERE THAT IS EQUIVALENT TO THE GLOBAL PRIOR VALUE
            # THE FACT IS THAT THE EFFECTIVE PRIOR HERE WILL BE THE SAME AS THE GENERATING FUNCTION (UP TO THE SNR PRIOR IF THAT IS CHANGED IN THE GENERATING FUNCTION)
            # generate_factors[:] += (self.xp.log(1 / (ends - starts)))[:, None]
            # generate_factors[:] += (self.xp.log(1 / (ends_fdot - starts_fdot)))[:, None]

        else:
            if current_priors is None:
                raise ValueError(
                    "If generating from the prior, must provide current_priors kwargs."
                )

            generated_points = current_priors.rvs(size=nwalkers * size).reshape(
                nwalkers, size, -1
            )
            if fill is not None or fill_inds is not None:
                if fill is None or fill_inds is None:
                    raise ValueError(
                        "If providing fill_inds or fill, must provide both."
                    )
                generated_points[fill_inds[0]] = fill.copy()

            generate_factors = current_priors.logpdf(
                generated_points.reshape(nwalkers * size, -1)
            ).reshape(nwalkers, size)

        """et = time.perf_counter()
        print("GENEARTE:", et - st)"""
        return generated_points, generate_factors

    def special_like_func(
        self,
        generated_points,
        base_shape,
        inds_reverse=None,
        old_d_h_d_h=None,
        overall_inds=None,
    ):
        # st = time.perf_counter()
        self.xp.cuda.runtime.setDevice(self.xp.cuda.runtime.getDevice())

        if overall_inds is None:
            raise ValueError("overall_inds is None.")

        # group everything

        # GENERATED POINTS MUST BE PASSED IN by reference not copied
        num_inds_change, nleaves_max, ndim = base_shape
        num_inds_change_gen, num_try, ndim_gen = generated_points.shape
        assert num_inds_change_gen == num_inds_change and ndim == ndim_gen

        if old_d_h_d_h is None:
            raise NotImplementedError
            self.d_h_d_h = d_h_d_h = (
                4 * self.df * self.xp.sum((in_vals.conj() * in_vals) / psd, axis=(1, 2))
            )
        else:
            self.d_h_d_h = d_h_d_h = self.xp.asarray(old_d_h_d_h)

        ll_out = self.xp.zeros((num_inds_change, num_try)).flatten()
        # TODO: take out of loop later?

        phase_marginalize = self.search
        generated_points_here = generated_points.reshape(-1, ndim)

        back_d_d = self.gb.d_d.copy()
        self.gb.d_d = self.xp.repeat(d_h_d_h, self.num_try)

        # do not need mapping because it comes in as overall inds already mapped
        data_index = self.xp.asarray(
            self.xp.repeat(overall_inds, self.num_try).astype(self.xp.int32)
        )

        noise_index = data_index.copy()

        self.data_index_check = data_index.reshape(-1, self.num_try)[:, 0]

        # TODO: Remove batch_size if GPU only ?
        prior_generated_points = generated_points_here

        if self.parameter_transforms is not None:
            prior_generated_points_in = self.parameter_transforms.both_transforms(
                prior_generated_points.copy(), xp=self.xp
            )

        N_temp = self.xp.asarray(
            get_N(
                prior_generated_points_in[:, 0],
                prior_generated_points_in[:, 1],
                self.waveform_kwargs["T"],
                self.waveform_kwargs["oversample"],
            )
        )

        waveform_kwargs_in = self.waveform_kwargs.copy()
        waveform_kwargs_in.pop("N")
        main_gpu = self.xp.cuda.runtime.getDevice()
        # TODO: do search sorted and apply that to nearest found for new points found with group

        # st3 = time.perf_counter()
        ll = self.gb.get_ll(
            prior_generated_points_in,
            self.mgh.data_list,
            self.mgh.psd_list,
            data_index=data_index,
            noise_index=noise_index,
            phase_marginalize=phase_marginalize,
            data_length=self.data_length,
            data_splits=self.mgh.gpu_splits,
            N=N_temp,
            return_cupy=True,
            **waveform_kwargs_in
        )
        self.xp.cuda.runtime.setDevice(main_gpu)
        self.xp.cuda.runtime.deviceSynchronize()

        """et3 = time.perf_counter()
        print("actual like:", et3 - st3)"""
        if self.xp.any(self.xp.isnan(ll)):
            assert self.xp.isnan(ll).sum() < 10
            ll[self.xp.isnan(ll)] = -1e300

        opt_snr = self.xp.sqrt(self.gb.h_h)

        if self.search:
            phase_maximized_snr = (
                self.xp.abs(self.gb.d_h) / self.xp.sqrt(self.gb.h_h)
            ).real.copy()

            phase_change = self.xp.angle(
                self.xp.asarray(self.gb.non_marg_d_h) / self.xp.sqrt(self.gb.h_h.real)
            )

            try:
                phase_maximized_snr = phase_maximized_snr.get()
                phase_change = phase_change.get()
                opt_snr = opt_snr.get()

            except AttributeError:
                pass

            if self.xp.any(self.xp.isnan(prior_generated_points)) or self.xp.any(
                self.xp.isnan(phase_change)
            ):
                breakpoint()

            # adjust for phase change from maximization
            generated_points_here[:, 3] = (
                generated_points_here[:, 3] - phase_change
            ) % (2 * np.pi)

            snr_comp = phase_maximized_snr

        else:
            snr_comp = (self.gb.d_h.real / self.xp.sqrt(self.gb.h_h)).real.copy()
            try:
                snr_comp = snr_comp.get()
                opt_snr = opt_snr.get()

            except AttributeError:
                pass

        snr_comp2 = snr_comp.reshape(-1, self.num_try)
        okay = snr_comp2 >= 1.0
        okay[inds_reverse.get()] = True
        ll[~okay.flatten()] = -1e300

        ##print(opt_snr[snr_comp.argmax()].real, snr_comp.max(), ll[snr_comp.argmax()].real - -1/2 * self.gb.d_d[snr_comp.argmax()].real)
        if self.search and self.search_snr_lim is not None:
            ll[
                (snr_comp < self.search_snr_lim * 0.8)
                | (opt_snr < self.search_snr_lim * self.search_snr_accept_factor)
            ] = -1e300
            """if self.xp.any(~((snr_comp
                < self.search_snr_lim * 0.95)
                | (opt_snr
                < self.search_snr_lim * self.search_snr_accept_factor))):
                breakpoint()"""

        generated_points[:] = generated_points_here.reshape(generated_points.shape)
        ll_out = ll.copy()
        # if inds_reverse is not None and len(inds_reverse) != 0 and split == num_splits - 1:
        #    breakpoint()

        ll_out = ll_out.reshape(num_inds_change, num_try)
        self.old_ll_out_check = ll_out.copy()
        if inds_reverse is not None:
            # try:
            #    tmp_d_h_d_h = d_h_d_h.get()
            # except AttributeError:
            #    tmp_d_h_d_h = d_h_d_h

            # this is special to GBs
            self.special_aux_ll = -ll_out[inds_reverse, 0]

            self.check_h_h = self.gb.h_h.reshape(-1, num_try)[inds_reverse]
            self.check_d_h = self.gb.d_h.reshape(-1, num_try)[inds_reverse]
            ll_out[inds_reverse, :] += self.special_aux_ll[:, None]

        # return gb.d_d
        self.gb.d_d = back_d_d.copy()
        # add noise term
        self.xp.cuda.runtime.deviceSynchronize()
        """et = time.perf_counter()
        print("LIKE:", et - st)"""
        return ll_out  #  + self.noise_ll[:, None]

    def special_prior_func(
        self, generated_points, base_shape, inds_reverse=None, **kwargs
    ):
        # st = time.perf_counter()
        nwalkers, nleaves_max, ndim = base_shape
        # st2 = time.perf_counter()
        lp_new = (
            self.gpu_priors["gb"]
            .logpdf(generated_points.reshape(-1, 8))
            .reshape(nwalkers, self.num_try)
        )
        """et2 = time.perf_counter()
        print("prior logpdf:", et2 - st2)"""
        lp_total = lp_new  # _old[:, None] + lp_new

        self.old_lp_total_check = lp_total.copy()
        if inds_reverse is not None:
            # this is special to GBs
            self.special_aux_lp = -lp_total[inds_reverse, 0]

            lp_total[inds_reverse, :] += self.special_aux_lp[:, None]

        # add noise lp
        """et = time.perf_counter()
        print("PRIOR:", et - st)"""
        return lp_total

    def readout_adjustment(self, out_vals, all_vals_prop, aux_all_vals, inds_reverse):
        self.out_vals, self.all_vals_prop, self.aux_all_vals, self.inds_reverse = (
            out_vals,
            all_vals_prop,
            aux_all_vals,
            inds_reverse,
        )

        (
            self.logP_out,
            self.ll_out,
            self.lp_out,
            self.log_proposal_pdf_out,
            self.log_sum_weights,
        ) = out_vals

        self.ll_out[inds_reverse] = self.special_aux_ll
        self.lp_out[inds_reverse] = self.special_aux_lp

    def get_proposal(
        self,
        gb_coords,
        inds,
        changes,
        leaf_inds_for_changes,
        band_inds,
        random,
        supps=None,
        branch_supps=None,
    ):
        """Make a proposal

        Args:
            all_coords (dict): Keys are ``branch_names``. Values are
                np.ndarray[ntemps, nwalkers, nleaves_max, ndim]. These are the curent
                coordinates for all the walkers.
            all_inds (dict): Keys are ``branch_names``. Values are
                np.ndarray[ntemps, nwalkers, nleaves_max]. These are the boolean
                arrays marking which leaves are currently used within each walker.
            all_inds_for_change (dict): Keys are ``branch_names``. Values are
                dictionaries. These dictionaries have keys ``"+1"`` and ``"-1"``,
                indicating waklkers that are adding or removing a leafm respectively.
                The values for these dicts are ``int`` np.ndarray[..., 3]. The "..." indicates
                the number of walkers in all temperatures that fall under either adding
                or removing a leaf. The second dimension, 3, is the indexes into
                the three-dimensional arrays within ``all_inds`` of the specific leaf
                that is being added or removed from those leaves currently considered.
            random (object): Current random state of the sampler.

        Returns:
            tuple: Tuple containing proposal information.
                First entry is the new coordinates as a dictionary with keys
                as ``branch_names`` and values as
                ``double `` np.ndarray[ntemps, nwalkers, nleaves_max, ndim] containing
                proposed coordinates. Second entry is the new ``inds`` array with
                boolean values flipped for added or removed sources. Third entry
                is the factors associated with the
                proposal necessary for detailed balance. This is effectively
                any term in the detailed balance fraction. +log of factors if
                in the numerator. -log of factors if in the denominator.

        """

        name = "gb"

        ntemps, nwalkers, nleaves_max, ndim = gb_coords.shape

        # adjust inds
        changes_gpu = self.xp.asarray(changes)
        # TODO: remove xp.asarrays here if they come in as GPU arrays
        inds_reverse_band = self.xp.where(changes_gpu == -1)

        leaf_inds_for_changes_gpu = self.xp.asarray(leaf_inds_for_changes)

        inds_reverse = inds_reverse_band[:2] + (
            leaf_inds_for_changes_gpu[inds_reverse_band],
        )
        inds_reverse_cpu = tuple([ir.get() for ir in list(inds_reverse)])
        inds_reverse_in = self.xp.where(changes_gpu.flatten() == -1)[0]
        inds_reverse_individual = leaf_inds_for_changes_gpu[changes_gpu < 0]

        inds_forward_band = self.xp.where(changes_gpu == +1)

        inds_forward = inds_forward_band[:2] + (
            leaf_inds_for_changes_gpu[inds_forward_band],
        )
        # inds_forward_in = self.xp.where(changes.flatten() == +1)[0]
        # inds_forward_individual = leaf_inds_for_changes_gpu[changes < 0]

        # add coordinates for new leaves
        current_priors = self.priors[name]

        inds_here_band = self.xp.where((changes_gpu == -1) | (changes_gpu == +1))

        inds_here = inds_here_band[:2] + (leaf_inds_for_changes_gpu[inds_here_band],)
        inds_here_cpu = tuple([ir.get() for ir in list(inds_here)])

        # allows for adjustment during search
        reset_num_try = False
        if len(inds_here[0]) != 0:
            if self.search and self.search_samples is not None:
                raise not NotImplementedError
                if len(self.search_inds) >= self.num_try:
                    reset_num_try = True
                    old_num_try = self.num_try
                self.num_try = len(self.search_inds)

            num_inds_change = len(inds_here[0])

            if self.provide_betas:  #  and not self.search:
                betas = self.xp.asarray(self.temperature_control.betas)[inds_here[0]]
            else:
                betas = None

            ll_here = self.xp.asarray(self.current_state.log_like.copy())[inds_here[:2]]
            lp_here = self.xp.asarray(self.current_state.log_prior)[inds_here[:2]]

            self.lp_old = lp_here

            rj_info = dict(
                ll=self.xp.zeros_like(ll_here), lp=self.xp.zeros_like(lp_here)
            )

            N_vals = branch_supps.holder["N_vals"]
            N_vals = self.xp.asarray(N_vals)

            self.inds_reverse = inds_reverse
            self.inds_forward = inds_forward

            if len(inds_reverse[0]) > 0:
                parameters_remove = self.parameter_transforms.both_transforms(
                    gb_coords[inds_reverse_cpu], xp=self.xp
                )

                group_index_tmp = inds_reverse[0] * nwalkers + inds_reverse[1]
                N_vals_in = self.xp.asarray(N_vals[inds_reverse])
                group_index = self.xp.asarray(
                    self.mgh.get_mapped_indices(group_index_tmp).astype(self.xp.int32)
                )

                waveform_kwargs_add = self.waveform_kwargs.copy()
                waveform_kwargs_add.pop("N")
                # removing these so d - (h - r) = d - h + r

                try:
                    self.xp.cuda.runtime.deviceSynchronize()
                    # parameters_remove[:, 0] *= 1e-30
                    self.gb.generate_global_template(
                        parameters_remove,
                        group_index,
                        self.mgh.data_list,
                        N=N_vals_in,
                        data_length=self.data_length,
                        data_splits=self.mgh.gpu_splits,
                        **waveform_kwargs_add
                    )
                except ValueError as e:
                    print(e)
                    breakpoint()

                # self.checkit4 =  (-1/2 * self.df * 4 * self.xp.sum(data_minus_template.conj() * data_minus_template / self.psd, axis=(2,3))) + self.xp.asarray(self.noise_ll).reshape(ntemps, nwalkers)

            overall_inds = supps.holder["overall_inds"][inds_here_cpu[:2]]
            base_shape = (len(inds_here[0]), 1, ndim)
            assert np.prod(band_inds.shape) == inds_here[0].shape[0]
            old_d_h_d_h = self.xp.zeros_like(ll_here)

            generate_points_out, logP_out, factors_out = self.get_mt_proposal(
                self.xp.asarray(gb_coords[inds_here_cpu]),
                len(inds_here[0]),
                inds_reverse_in,
                random,
                args_prior=(base_shape, inds_reverse_in),
                kwargs_generate={
                    "current_priors": current_priors,
                    "band_inds": band_inds.flatten(),
                },
                args_like=(base_shape,),
                kwargs_like={
                    "inds_reverse": inds_reverse_in,
                    "old_d_h_d_h": old_d_h_d_h,
                    "overall_inds": overall_inds,
                },
                betas=betas,
                rj_info=rj_info,
            )

            # gb_coords[inds_forward] = generate_points_out.copy()

            # TODO: make sure detailed balance this will move to detailed balance in multiple try

            self.logP_out = logP_out

            return (
                generate_points_out,
                self.ll_out,
                self.lp_out,
                factors_out,
                betas,
                inds_here,
            )

        else:
            breakpoint()

    def propose(self, model, state):
        """Use the move to generate a proposal and compute the acceptance

        Args:
            model (:class:`eryn.model.Model`): Carrier of sampler information.
            state (:class:`State`): Current state of the sampler.

        Returns:
            :class:`State`: State of sampler after proposal is complete.

        """

        st = time.perf_counter()
        # TODO: keep this?
        # this exposes anywhere in the proposal class to this information
        self.current_state = state
        self.current_model = model

        # Run any move-specific setup.
        self.setup(state.branches)

        # ll_before = model.compute_log_like_fn(state.branches_coords, inds=state.branches_inds, supps=state.supplimental, branch_supps=state.branches_supplimental)

        # if not np.allclose(ll_before[0], state.log_like):
        #    breakpoint()

        new_state = State(state, copy=True)
        all_branch_names = list(new_state.branches_coords.keys())

        if np.any(
            new_state.branches_supplimental["gb"].holder["N_vals"][
                new_state.branches_inds["gb"]
            ]
            == 0
        ):
            breakpoint()

        self.mgh.map = new_state.supplimental.holder["overall_inds"].flatten()

        ntemps, nwalkers, _, _ = new_state.branches[
            list(new_state.branches.keys())[0]
        ].shape

        num_consecutive_rj_moves = 1
        # print("starting")

        for rj_move_i in range(num_consecutive_rj_moves):
            # st = time.perf_counter()
            accepted = np.zeros((ntemps, nwalkers), dtype=bool)

            coords_propose_in = self.xp.asarray(new_state.branches_coords["gb"])
            inds_propose_in = self.xp.asarray(new_state.branches_inds["gb"])
            branches_supp_propose_in = new_state.branches_supplimental["gb"]
            remaining_coords = coords_propose_in[inds_propose_in]
            f0 = remaining_coords[:, 1] / 1e3
            inds_into_full_f0 = self.xp.arange(f0.shape[0])

            band_inds = self.xp.searchsorted(self.band_edges, f0) - 1
            temp_inds = self.xp.repeat(
                self.xp.arange(ntemps)[:, None],
                np.prod(coords_propose_in.shape[1:3]),
                axis=1,
            ).reshape(ntemps, nwalkers, -1)[inds_propose_in]
            walkers_inds = self.xp.repeat(
                self.xp.repeat(self.xp.arange(nwalkers)[None, :], ntemps, axis=0)[
                    :, :, None
                ],
                coords_propose_in.shape[2],
                axis=2,
            )[inds_propose_in]
            leaf_inds = self.xp.repeat(
                self.xp.arange(coords_propose_in.shape[2])[None, :],
                ntemps * nwalkers,
                axis=0,
            ).reshape(ntemps, nwalkers, -1)[inds_propose_in]

            # TODO: make this weighted somehow to not focus on empty bands
            # we are already leaving out the last one
            bands_per_walker = self.xp.tile(
                (self.xp.arange(self.num_bands - 1)), (ntemps, nwalkers, 1)
            )

            """et = time.perf_counter()
            print("initial setup", et - st)"""
            # st = time.perf_counter()

            # odds & evens (evens first)
            for i in range(2):
                # st = time.perf_counter()
                bands_per_walker_here = bands_per_walker[
                    bands_per_walker % 2 == i
                ].reshape(ntemps, nwalkers, -1)
                max_band_ind = bands_per_walker_here.max().item()
                band_inds_here = (
                    int(1e12) * temp_inds[band_inds % 2 == i]
                    + int(1e6) * walkers_inds[band_inds % 2 == i]
                    + band_inds[band_inds % 2 == i]
                )

                inds_into_full_f0_here = inds_into_full_f0[band_inds.flatten() % 2 == i]

                inds_for_shuffle = self.xp.arange(len(inds_into_full_f0_here))
                # shuffle allows the first unique index to be the one that is dropped
                self.xp.random.shuffle(inds_for_shuffle)
                bands_inds_here_tmp = band_inds_here[inds_for_shuffle]

                assert band_inds_here.shape == inds_into_full_f0_here.shape

                # shuffle and then take first index from band
                # is a good way to get random binary from that band
                (
                    unique_bands_inds_here,
                    unique_bands_inds_here_index,
                    unique_bands_inds_here_inverse,
                    unique_bands_inds_here_counts,
                ) = self.xp.unique(
                    bands_inds_here_tmp,
                    return_index=True,
                    return_counts=True,
                    return_inverse=True,
                )

                # selected_inds_from_shuffle is indices reverted back to
                # inds_into_full_f0_here after shuffle
                selected_inds_from_shuffle = inds_for_shuffle[
                    unique_bands_inds_here_index
                ]

                # frequencies of the first instance found in band according to self.xp.unique call
                # and what bands they belong to
                inds_into_full_f0_here_selected = inds_into_full_f0_here[
                    selected_inds_from_shuffle
                ]
                bands_inds_here_selected = band_inds_here[selected_inds_from_shuffle]

                temp_index_count = temp_inds[inds_into_full_f0_here_selected]
                walkers_index_count = walkers_inds[inds_into_full_f0_here_selected]
                leaf_index_count = leaf_inds[inds_into_full_f0_here_selected]

                # get band index for each selected source by maths
                band_index_count = self.xp.floor(
                    (
                        unique_bands_inds_here
                        - temp_index_count * int(1e12)
                        - walkers_index_count * int(1e6)
                    )
                ).astype(int)
                band_index_tmp = self.xp.searchsorted(
                    bands_per_walker_here[0, 0], band_index_count, side="left"
                )

                nleaves_here = self.xp.zeros_like(bands_per_walker_here)

                # how many leaves are in each band
                try:
                    nleaves_here[
                        (temp_index_count, walkers_index_count, band_index_tmp)
                    ] = unique_bands_inds_here_counts
                except IndexError:
                    breakpoint()

                # setup leaf count change arrays
                if self.fix_change is None:
                    changes = self.xp.random.choice(
                        [-1, 1], size=bands_per_walker_here.shape
                    )
                else:
                    changes = self.xp.full(bands_per_walker_here.shape, self.fix_change)

                # make sure to add a binary if there are None
                changes[nleaves_here == 0] = +1

                if self.xp.any(
                    nleaves_here >= self.max_k[all_branch_names.index("gb")]
                ):
                    raise ValueError("nleaves_here higher than max_k.")

                # number of sub bands
                num_sub_bands_here = bands_per_walker_here.shape[-1]

                # build arrays to help properly index and locate within these sub bands
                temp_inds_for_change = self.xp.repeat(
                    self.xp.arange(ntemps)[:, None],
                    nwalkers * num_sub_bands_here,
                    axis=-1,
                ).reshape(ntemps, nwalkers, num_sub_bands_here)
                walker_inds_for_change = self.xp.tile(
                    self.xp.arange(nwalkers), (ntemps, num_sub_bands_here, 1)
                ).transpose((0, 2, 1))
                band_inds_for_change = self.xp.tile(
                    self.xp.arange(num_sub_bands_here), (ntemps, nwalkers, 1)
                )

                # leaves will be filled later
                leaf_inds_for_change = -self.xp.ones(
                    (ntemps, nwalkers, num_sub_bands_here), dtype=int
                )

                # add leaves that would be removed if count is not currently zero
                # will adjust for count soon
                # len(leaf_index_count) < total number of proposals
                leaf_inds_for_change[
                    (temp_index_count, walkers_index_count, band_index_tmp)
                ] = leaf_index_count

                # band_inds_for_change[(temp_index_count, walkers_index_count, band_index_tmp)] = band_index_tmp

                leaf_inds_tmp = self.xp.repeat(
                    self.xp.arange(coords_propose_in.shape[2])[None, :],
                    ntemps * nwalkers,
                    axis=0,
                ).reshape(ntemps, nwalkers, -1)

                # any binary that exists give fake high value to remove from sort
                leaf_inds_tmp[inds_propose_in] = int(1e7)

                # this gives unused leaves, you take the number you need for each walker / temperature
                leaf_inds_tmp_2 = self.xp.sort(leaf_inds_tmp, axis=-1)[
                    :, :, :num_sub_bands_here
                ]
                leaf_inds_for_change[changes > 0] = leaf_inds_tmp_2[changes > 0]

                assert not self.xp.any(leaf_inds_for_change == -1)

                if self.xp.any(leaf_inds_tmp_2 == int(1e7)):
                    raise ValueError(
                        "Not enough spots to allocate for new binaries. Need to increase max leaves."
                    )

                # TODO: check that number of available spots is high enough

                """et = time.perf_counter()
                print("start", et - st)
                st = time.perf_counter()"""
                # propose new sources and coordinates
                (
                    new_coords,
                    ll_out,
                    lp_out,
                    factors,
                    betas,
                    inds_here,
                ) = self.get_proposal(
                    coords_propose_in,
                    inds_propose_in,
                    changes,
                    leaf_inds_for_change,
                    bands_per_walker_here,
                    model.random,
                    branch_supps=branches_supp_propose_in,
                    supps=new_state.supplimental,
                )

                """et = time.perf_counter()
                print("proposal", et - st)
                st = time.perf_counter()"""

                # TODO: check this
                edge_factors = self.xp.zeros_like(factors)
                # get factors for edges
                min_k = self.min_k[all_branch_names.index("gb")]
                max_k = self.max_k[all_branch_names.index("gb")]

                # fix proposal asymmetry at bottom of k range
                inds_min = self.xp.where(nleaves_here.flatten() == min_k)
                # numerator term so +ln
                edge_factors[inds_min] += np.log(1 / 2.0)

                # fix proposal asymmetry at top of k range
                inds_max = self.xp.where(nleaves_here.flatten() == max_k)
                # numerator term so -ln
                edge_factors[inds_max] += np.log(1 / 2.0)

                # fix proposal asymmetry at bottom of k range (kmin + 1)
                inds_min = self.xp.where(nleaves_here.flatten() == min_k + 1)
                # numerator term so +ln
                edge_factors[inds_min] -= np.log(1 / 2.0)

                # fix proposal asymmetry at top of k range (kmax - 1)
                inds_max = self.xp.where(nleaves_here.flatten() == max_k - 1)
                # numerator term so -ln
                edge_factors[inds_max] -= np.log(1 / 2.0)

                factors += edge_factors

                prev_logl = self.xp.asarray(new_state.log_like)[inds_here[:2]]

                prev_logp = self.xp.asarray(new_state.log_prior)[inds_here[:2]]

                """et = time.perf_counter()
                print("prior", et - st)
                st = time.perf_counter()"""
                logl = prev_logl + ll_out

                logp = prev_logp + lp_out
                # loglcheck, new_blobs = model.compute_log_like_fn(q, inds=new_inds, logp=logp, supps=new_supps, branch_supps=new_branch_supps)
                # if not self.xp.all(self.xp.abs(logl[logl != -1e300] - loglcheck[logl != -1e300]) < 1e-5):
                #    breakpoint()

                logP = self.compute_log_posterior(logl, logp, betas=betas)

                # TODO: check about prior = - inf
                # takes care of tempering
                prev_logP = self.compute_log_posterior(
                    prev_logl, prev_logp, betas=betas
                )

                # TODO: fix this
                # this is where _metropolisk should come in
                lnpdiff = factors + logP - prev_logP

                accepted = lnpdiff > self.xp.log(
                    self.xp.asarray(model.random.rand(*lnpdiff.shape))
                )

                """et = time.perf_counter()
                print("through accepted", et - st)
                st = time.perf_counter()"""

                # bookkeeping

                inds_reverse = self.xp.where(changes.flatten() == -1)[0]

                # adjust births from False -> True
                inds_forward = self.xp.where(changes.flatten() == +1)[0]

                # accepted_keep = (temp_inds_for_change.flatten()[inds_forward] == 0) & (walker_inds_for_change.flatten()[inds_forward] == 1)

                """accepted_keep_tmp = self.xp.where(accepted_keep)[0]
                accepted_keep[accepted_keep_tmp[:]] = False
                accepted_keep[accepted_keep_tmp[164679]] = True"""

                # accepted[:] = False
                # accepted[inds_forward] = True
                """accepted[inds_forward[:]] = False
                accepted[inds_reverse[0]] = False"""

                # not accepted removals
                accepted_reverse = accepted[inds_reverse]
                accepted_inds_reverse = inds_reverse[accepted_reverse]
                not_accepted_inds_reverse = inds_reverse[~accepted_reverse]

                tuple_not_accepted_reverse = (
                    temp_inds_for_change.flatten()[not_accepted_inds_reverse],
                    walker_inds_for_change.flatten()[not_accepted_inds_reverse],
                    leaf_inds_for_change.flatten()[not_accepted_inds_reverse],
                )

                tuple_accepted_reverse = (
                    temp_inds_for_change.flatten()[accepted_inds_reverse],
                    walker_inds_for_change.flatten()[accepted_inds_reverse],
                    leaf_inds_for_change.flatten()[accepted_inds_reverse],
                )

                if len(not_accepted_inds_reverse) > 0:
                    points_not_accepted_removal = coords_propose_in[
                        tuple_not_accepted_reverse
                    ]

                else:
                    points_not_accepted_removal = self.xp.empty((0, 8))

                tuple_accepted_reverse_cpu = tuple(
                    [tmp.get() for tmp in list(tuple_accepted_reverse)]
                )
                # accepted removals
                new_state.branches["gb"].inds[tuple_accepted_reverse_cpu] = False
                delta_logl_trans1 = self.xp.zeros_like(
                    leaf_inds_for_change, dtype=float
                )
                delta_logl_trans1[
                    (
                        temp_inds_for_change.flatten()[accepted_inds_reverse],
                        walker_inds_for_change.flatten()[accepted_inds_reverse],
                        band_inds_for_change.flatten()[accepted_inds_reverse],
                    )
                ] = (
                    logl[accepted_inds_reverse] - prev_logl[accepted_inds_reverse]
                )

                new_state.log_like += delta_logl_trans1.sum(axis=-1).get()

                # accepted update logp
                delta_logp_trans = self.xp.zeros_like(leaf_inds_for_change, dtype=float)
                delta_logp_trans[
                    (
                        temp_inds_for_change.flatten()[accepted_inds_reverse],
                        walker_inds_for_change.flatten()[accepted_inds_reverse],
                        band_inds_for_change.flatten()[accepted_inds_reverse],
                    )
                ] = (
                    logp[accepted_inds_reverse] - prev_logp[accepted_inds_reverse]
                )

                new_state.log_prior += delta_logp_trans.sum(axis=-1).get()

                # accepted additions
                accepted_forward = accepted[inds_forward]
                accepted_inds_forward = inds_forward[accepted_forward]
                not_accepted_inds_forward = inds_forward[~accepted_forward]

                tuple_accepted_forward = (
                    temp_inds_for_change.flatten()[accepted_inds_forward],
                    walker_inds_for_change.flatten()[accepted_inds_forward],
                    leaf_inds_for_change.flatten()[accepted_inds_forward],
                )

                tuple_accepted_forward_cpu = tuple(
                    [tmp.get() for tmp in list(tuple_accepted_forward)]
                )
                new_state.branches["gb"].inds[tuple_accepted_forward_cpu] = True
                new_state.branches["gb"].coords[
                    tuple_accepted_forward_cpu
                ] = new_coords[accepted_forward].get()

                """et = time.perf_counter()
                print("bookkeeping", et - st)
                st = time.perf_counter()"""

                if len(accepted_inds_forward) > 0:
                    points_accepted_addition = new_coords[accepted_forward]

                    # get group friend finder information
                    f0_accepted_addition = -100 * self.xp.ones_like(
                        leaf_inds_for_change, dtype=float
                    )
                    f0_accepted_addition[
                        (
                            temp_inds_for_change.flatten()[accepted_inds_forward],
                            walker_inds_for_change.flatten()[accepted_inds_forward],
                            band_inds_for_change.flatten()[accepted_inds_forward],
                        )
                    ] = (
                        points_accepted_addition[:, 1] / 1e3
                    )

                    for t in range(ntemps):
                        # use old state to get supp information
                        f0_old = (
                            self.xp.asarray(
                                state.branches["gb"].coords[
                                    t, state.branches["gb"].inds[t]
                                ][:, 1]
                            )
                            / 1e3
                        )
                        friend_start_inds = self.xp.asarray(
                            state.branches["gb"].branch_supplimental.holder[
                                "friend_start_inds"
                            ][t, state.branches["gb"].inds[t]]
                        )

                        f0_old_sorted = self.xp.sort(f0_old, axis=-1)
                        inds_f0_old_sorted = self.xp.argsort(f0_old, axis=-1)
                        f0_accepted_addition_in1 = f0_accepted_addition[t]
                        f0_accepted_addition_in2 = f0_accepted_addition_in1[
                            f0_accepted_addition_in1 > -1.0
                        ]
                        temp_inds_in = temp_inds_for_change[t][
                            f0_accepted_addition_in1 > -1.0
                        ]
                        walker_inds_in = walker_inds_for_change[t][
                            f0_accepted_addition_in1 > -1.0
                        ]
                        leaf_inds_in = leaf_inds_for_change[t][
                            f0_accepted_addition_in1 > -1.0
                        ]

                        inds_f0_accepted_addition = (
                            self.xp.searchsorted(
                                f0_old_sorted, f0_accepted_addition_in2, side="right"
                            )
                            - 1
                        )

                        old_inds_f0_old = inds_f0_old_sorted[inds_f0_accepted_addition]
                        comp_f0_old = f0_old[old_inds_f0_old]
                        new_friends_start_inds = friend_start_inds[old_inds_f0_old]

                        # TODO: maybe check this
                        new_state.branches["gb"].branch_supplimental.holder[
                            "friend_start_inds"
                        ][
                            (
                                temp_inds_in.get(),
                                walker_inds_in.get(),
                                leaf_inds_in.get(),
                            )
                        ] = new_friends_start_inds.get()

                    """inds_into_old_array = self.xp.take_along_axis(inds_f0_old_sorted, inds_f0_accepted_addition, axis=-1).reshape(ntemps, nwalkers, -1)

                    new_start_inds = self.xp.take_along_axis(state.branches["gb"].branch_supplimental.holder["friend_start_inds"], inds_into_old_array, axis=-1)

                    keep_new_start_inds = new_start_inds[f0_accepted_addition.reshape(ntemps, nwalkers, -1) > -10.0]
                    breakpoint()
                    state.branches["gb"].branch_supplimental.holder["friend_start_inds"][(

                    )] = keep_new_start_inds
                    
                    inds_into_old_array 
                    keep_inds_f0_accepted_addition = inds_f0_accepted_addition > 0



                    inds_into_old_array = inds_f0_old_sorted[inds_f0_accepted_addition]"""

                else:
                    points_accepted_addition = self.xp.empty((0, 8))

                delta_logl_trans2 = self.xp.zeros_like(
                    leaf_inds_for_change, dtype=float
                )
                delta_logl_trans2[
                    (
                        temp_inds_for_change.flatten()[accepted_inds_forward],
                        walker_inds_for_change.flatten()[accepted_inds_forward],
                        band_inds_for_change.flatten()[accepted_inds_forward],
                    )
                ] = (
                    logl[accepted_inds_forward] - prev_logl[accepted_inds_forward]
                )

                new_state.log_like += delta_logl_trans2.sum(axis=-1).get()

                delta_logp_trans = self.xp.zeros_like(leaf_inds_for_change, dtype=float)
                delta_logp_trans[
                    (
                        temp_inds_for_change.flatten()[accepted_inds_forward],
                        walker_inds_for_change.flatten()[accepted_inds_forward],
                        band_inds_for_change.flatten()[accepted_inds_forward],
                    )
                ] = (
                    logp[accepted_inds_forward] - prev_logp[accepted_inds_forward]
                )

                new_state.log_prior += delta_logp_trans.sum(axis=-1).get()

                accepted_counts = self.xp.zeros_like(leaf_inds_for_change, dtype=bool)
                num_proposals_here = leaf_inds_for_change.shape[-1]

                accepted_counts[
                    (
                        temp_inds_for_change.flatten()[accepted],
                        walker_inds_for_change.flatten()[accepted],
                        band_inds_for_change.flatten()[accepted],
                    )
                ] = True

                accepted_overall = accepted_counts.sum(axis=-1)

                # TODO: if we do not run every band, we need to adjust this
                self.accepted += accepted_overall.get()
                self.num_proposals += num_proposals_here

                points_to_add_to_template = self.xp.concatenate(
                    [
                        points_accepted_addition,
                        points_not_accepted_removal,  # were removed at the start, need to be put back
                    ],
                    axis=0,
                )

                """et = time.perf_counter()
                print("before add", et - st)
                st = time.perf_counter()"""

                if points_to_add_to_template.shape[0] > 0:
                    points_to_add_to_template_in = (
                        self.parameter_transforms.both_transforms(
                            points_to_add_to_template, xp=self.xp
                        )
                    )
                    N_temp = get_N(
                        points_to_add_to_template_in[:, 0],
                        points_to_add_to_template_in[:, 1],
                        self.waveform_kwargs["T"],
                        self.waveform_kwargs["oversample"],
                    )

                    # has to be accepted forward first (sase as points_to_add_to_template)
                    inds_here_for_add = tuple(
                        [
                            self.xp.concatenate(
                                [
                                    tuple_accepted_forward[jjj],
                                    tuple_not_accepted_reverse[jjj],
                                ]
                            )
                            for jjj in range(len(tuple_not_accepted_reverse))
                        ]
                    )

                    inds_here_for_add_cpu = tuple(
                        [tmp.get() for tmp in list(inds_here_for_add)]
                    )

                    # inds_here_for_add = tuple_accepted_forward

                    new_state.branches["gb"].branch_supplimental.holder["N_vals"][
                        inds_here_for_add_cpu
                    ] = N_temp.get()

                    group_index_tmp_accepted_forward = (
                        tuple_accepted_forward[0] * nwalkers + tuple_accepted_forward[1]
                    )
                    group_index_tmp_not_accepted_reverse = (
                        tuple_not_accepted_reverse[0] * nwalkers
                        + tuple_not_accepted_reverse[1]
                    )

                    group_index_tmp = self.xp.concatenate(
                        [
                            group_index_tmp_accepted_forward,
                            group_index_tmp_not_accepted_reverse,
                        ]
                    )
                    group_index = self.xp.asarray(
                        self.mgh.get_mapped_indices(group_index_tmp).astype(
                            self.xp.int32
                        )
                    )

                    # adding these so d - (h + a) = d - h - a
                    factors_multiply = -self.xp.ones_like(N_temp, dtype=self.xp.float64)
                    waveform_kwargs_add = self.waveform_kwargs.copy()
                    waveform_kwargs_add.pop("N")

                    # points_to_add_to_template_in[1:, 0] *= 1e-30

                    self.gb.generate_global_template(
                        points_to_add_to_template_in,
                        group_index,
                        self.mgh.data_list,
                        N=N_temp,
                        data_length=self.data_length,
                        data_splits=self.mgh.gpu_splits,
                        factors=factors_multiply,
                        **waveform_kwargs_add
                    )

                    if np.any(
                        new_state.branches_supplimental["gb"].holder["N_vals"][
                            new_state.branches_inds["gb"]
                        ]
                        == 0
                    ):
                        breakpoint()

                    """et = time.perf_counter()
                    print("after add", et - st)
                    st = time.perf_counter()"""

        # st = time.perf_counter()
        if self.time % 1 == 0:
            ll_after2 = (
                self.mgh.get_ll(include_psd_info=True)
                .flatten()[new_state.supplimental[:]["overall_inds"]]
                .reshape(ntemps, nwalkers)
            )
            if np.abs(ll_after2 - new_state.log_like).max() > 1e-4:
                if np.abs(ll_after2 - new_state.log_like).max() > 10.0:
                    breakpoint()
                fix = np.abs(ll_after2 - new_state.log_like) > 1e-4
                new_state.log_like[fix] = ll_after2[fix]

        # print("rj", (et - st) / num_consecutive_rj_moves)

        if False:  # self.temperature_control is not None and not self.prevent_swaps:
            # TODO: add swaps?
            new_state = self.temperature_control.temper_comps(new_state, adapt=False)

        # et = time.perf_counter()
        # print("swapping", et - st)
        self.mgh.map = new_state.supplimental.holder["overall_inds"].flatten()
        accepted = np.zeros_like(new_state.log_like)
        self.time += 1

        self.mempool.free_all_blocks()

        et = time.perf_counter()
        print("RJ end", et - st)
        print(new_state.branches["gb"].nleaves.mean(axis=-1))
        return new_state, accepted



from lisaconstants import *
import numpy as np

co = C_SI


def HeavisideTheta(x, xp=None):
    if xp is None:
        xp = np

    squeeze = True if x.ndim == 1 else False

    x = xp.atleast_1d(x)

    out = 1.0 * (x >= 0.0)

    if squeeze:
        out = out.squeeze()

    return out


def DiracDelta(x, xp=None):
    if xp is None:
        xp = np
    squeeze = True if x.ndim == 1 else False

    x = xp.atleast_1d(x)

    # 1e-10 to guard against numerical error
    out = 1.0 * (xp.abs(x) < 1e-10)

    if squeeze:
        out = out.squeeze()

    return out


tau_2_default = 5585.708541201614
Deltav_default = 1.1079114425597559 * 10 ** (-9)

tau_2_default = 1.9394221536001746
Deltav_default = 2.22616837 * 10 ** (-11)
t0 = 600


def tdi_glitch_XYZ1(
    t_in, T=8.3, tau_2=tau_2_default, Deltav=Deltav_default, t0=600, mtm=1.982, xp=None
):
    # def tdi_glitch_XYZ1(t_in, T=8.3, tau_1=480.0, tau_2=100.0, Deltav=1e-12, t0=600.0, mtm=1.982, xp=None):

    if xp is None:
        xp = np

    tau_2 = xp.atleast_1d(tau_2)
    t0 = xp.atleast_1d(t0)
    Deltav = xp.atleast_1d(Deltav)

    assert tau_2.shape == t0.shape == Deltav.shape

    out = xp.zeros((3, len(t0), len(t_in)))

    run = ~(
        xp.isinf(xp.exp((-t_in[None, :] + t0[:, None])))
        | xp.isnan(xp.exp((-t_in[None, :] + t0[:, None])))
    )

    tdiX1link12 = (
        mtm
        * Deltav[:, None]
        * (
            t_in[None, :]
            - t0[:, None]
            - 2 * tau_2[:, None]
            + xp.exp((-t_in[None, :] + t0[:, None]) / tau_2[:, None])
            * (t_in[None, :] - t0[:, None] + 2 * tau_2[:, None])
        )
        * DiracDelta(t_in[None, :] - t0[:, None])
        + mtm
        * Deltav[:, None]
        * (
            1
            + (
                xp.exp((-t_in[None, :] + t0[:, None]) / tau_2[:, None])
                * (-t_in[None, :] + t0[:, None] - tau_2[:, None])
            )
            / tau_2[:, None]
        )
        * HeavisideTheta(t_in[None, :] - t0[:, None])
    ) / co - (
        mtm
        * Deltav[:, None]
        * (
            t_in[None, :]
            + 4 * T
            - t0[:, None]
            - 2 * tau_2[:, None]
            + xp.exp((-t_in[None, :] - 4 * T + t0[:, None]) / tau_2[:, None])
            * (t_in[None, :] + 4 * T - t0[:, None] + 2 * tau_2[:, None])
        )
        * DiracDelta(t_in[None, :] + 4 * T - t0[:, None])
        + mtm
        * Deltav[:, None]
        * (
            1
            - (
                xp.exp((-t_in[None, :] - 4 * T + t0[:, None]) / tau_2[:, None])
                * (t_in[None, :] + 4 * T - t0[:, None] + tau_2[:, None])
            )
            / tau_2[:, None]
        )
        * HeavisideTheta(t_in[None, :] + 4 * T - t0[:, None])
    ) / co

    tdiY1link12 = (
        -2
        * (
            mtm
            * Deltav[:, None]
            * (
                t_in[None, :]
                + T
                - t0[:, None]
                - 2 * tau_2[:, None]
                + (t_in[None, :] + T - t0[:, None] + 2 * tau_2[:, None])
                / xp.exp((t_in[None, :] + T - t0[:, None]) / tau_2[:, None])
            )
            * DiracDelta(t_in[None, :] + T - t0[:, None])
            + mtm
            * Deltav[:, None]
            * (
                1
                - (t_in[None, :] + T - t0[:, None] + tau_2[:, None])
                / (
                    xp.exp((t_in[None, :] + T - t0[:, None]) / tau_2[:, None])
                    * tau_2[:, None]
                )
            )
            * HeavisideTheta(t_in[None, :] + T - t0[:, None])
        )
    ) / co + (
        2
        * (
            mtm
            * Deltav[:, None]
            * (
                t_in[None, :]
                + 3 * T
                - t0[:, None]
                - 2 * tau_2[:, None]
                + xp.exp((-t_in[None, :] - 3 * T + t0[:, None]) / tau_2[:, None])
                * (t_in[None, :] + 3 * T - t0[:, None] + 2 * tau_2[:, None])
            )
            * DiracDelta(t_in[None, :] + 3 * T - t0[:, None])
            + mtm
            * Deltav[:, None]
            * (
                1
                - (
                    xp.exp((-t_in[None, :] - 3 * T + t0[:, None]) / tau_2[:, None])
                    * (t_in[None, :] + 3 * T - t0[:, None] + tau_2[:, None])
                )
                / tau_2[:, None]
            )
            * HeavisideTheta(t_in[None, :] + 3 * T - t0[:, None])
        )
    ) / co

    tdiZ1link12 = xp.zeros_like(tdiX1link12)

    tdiX1link12[~run] = 0.0
    tdiY1link12[~run] = 0.0
    tdiZ1link12[~run] = 0.0

    out[0, run] = tdiX1link12[run]
    out[1, run] = tdiY1link12[run]
    out[2, run] = tdiZ1link12[run]

    return out


if __name__ == "__main__":
    dt = 1 / 0.3
    t = np.arange(2e3) * dt
    check = tdi_glitch_XYZ1(
        t
    )  # , T=8.3, tau_1=480.0, tau_2=100.0, Deltav=1e-12, t0=600.0, mtm=1.982)
    import matplotlib.pyplot as plt

    plt.loglog(np.fft.rfftfreq(len(check[0]), dt), np.abs(np.fft.rfft(check[0])))
    plt.loglog(np.fft.rfftfreq(len(check[0]), dt), np.abs(np.fft.rfft(check[1])))
    plt.savefig("plot1.png")
    breakpoint()
